// {{{ copyright

/********************************************************************
 *
 * $Id: RuntimeCommandException.java,v 1.3 2000/11/08 11:39:30 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.command;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A RuntimeCommandException is useful to pass RuntimeExceptions that were
 * expected and can be handled up to the CommandDistributor, without having to
 * declare it. <p>
 *
 * Otherwise the RuntimeCommandException has the same semantic as a {@link
 * MessageCommandException MessageCommandException} except for an additional
 * nested exception that is the original source of the error.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.3 $
 * @since       0.98.0
 */
public class RuntimeCommandException
    extends RuntimeException
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger (RuntimeCommandException.class);

    /**
     * The nested exception.
     */
    private Exception nested;

    /**
     * A parent component for the message display.
     */
    private Object parent;

    /**
     * The messageResource of the RuntimeCommandException.
     */
    private String messageResource;

    /**
     * The messageParameters of the RuntimeCommandException.
     */
    private Object[] messageParameters;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ RuntimeCommandException(Exception)

    /**
     * Create a new RuntimeCommandException.
     *
     * @param   nested The original source of the error.
     */
    public RuntimeCommandException (Exception nested)
    {
        super(nested == null ? "Unknown" : nested.getMessage());
        this.nested = nested;
    }


    // }}}
    // {{{ RuntimeCommandException(Exception,String)

    /**
     * Create a new RuntimeCommandException.
     *
     * @param   nested The original source of the error.
     * @param   message A message detailing the reason for the exception.
     */
    public RuntimeCommandException (Exception nested, String message)
    {
        this(nested, message, null, null, null);
    }


    // }}}
    // {{{ RuntimeCommandException(Exception,String,Object,String)

    /**
     * Create a new RuntimeCommandException.
     *
     * @param   nested The original source of the error.
     * @param   message A message detailing the reason for the exception.
     * @param   parent  The parent Component for the dialog to display.
     * @param   messageResource The resource basename for the message dialog.
     *
     * @see     de.qfs.lib.gui.Message#showMessage(Component,String,Object[])
     */
    public RuntimeCommandException (Exception nested, String message,
                                    Object parent, String messageResource)

    {
        this (nested, messageResource, parent, messageResource, null);
    }


    // }}}
    // {{{ RuntimeCommandException(Exception,String,Object,String,Object[])

    /**
     * Create a new RuntimeCommandException.
     *
     * @param   nested The original source of the error.
     * @param   message A message detailing the reason for the exception.
     * @param   parent  The parent Component for the dialog to display.
     * @param   messageResource The resource basename for the message dialog.
     * @param   parameters The parameters for the message.
     *
     * @see     de.qfs.lib.gui.Message#showMessage(Component,String,Object[])
     */
    public RuntimeCommandException (Exception nested, String message,
                                    Object parent, String messageResource,
                                    Object[] parameters)

    {
        super(message);
        this.nested = nested;
        this.parent = parent;
        this.messageResource = messageResource;
        this.messageParameters = parameters;
    }


    // }}}

    //----------------------------------------------------------------------
    // Member access
    //----------------------------------------------------------------------
    // {{{ getNested

    /**
     * Get the original reason for the exception.
     *
     * @return  The nested Exception.
     */
    public Exception getNested()
    {
        return nested;
    }

    // }}}
    // {{{ getParent

    /**
     * Get the parent of the RuntimeCommandException.
     *
     * @return  The parent of the RuntimeCommandException.
     */
    public final Object getParent()
    {
        return parent;
    }

    // }}}
    // {{{ setParent

    /**
     * Set the target component for the message display. This is useful since
     * that way a nested library method doesn't have to know about the GUI
     * context from which it was called.
     *
     * @param   parent  The new parent to set.
     */
    public void setParent(Object parent)
    {
        this.parent = parent;
    }

    // }}}
    // {{{ getMessageResource

    /**
     * Get the messageResource of the RuntimeCommandException.
     *
     * @return  The messageResource of the RuntimeCommandException.
     */
    public final String getMessageResource()
    {
        return messageResource;
    }

    // }}}
    // {{{ setMessageResource

    /**
     * Set the messageResource for the RuntimeCommandException. The original
     * caller of the method that caused the Exception may have additional
     * information that makes this necessary.
     *
     *
     * @param   messageResource The new messageResource.
     */
    public void setMessageResource(String messageResource)
    {
        this.messageResource = messageResource;
    }

    // }}}
    // {{{ getMessageParameters

    /**
     * Get the messageParameters of the RuntimeCommandException.
     *
     * @return  The messageParameters of the RuntimeCommandException.
     */
    public final Object[] getMessageParameters()
    {
        return messageParameters;
    }

    // }}}
    // {{{ setMessageParameters

    /**
     * Set the messageParameters for the RuntimeCommandException. The original
     * caller of the method that caused the Exception may have additional
     * information that makes this necessary.
     *
     *
     * @param   messageParameters The new messageParameters.
     */
    public void setMessageParameters(Object[] messageParameters)
    {
        this.messageParameters = messageParameters;
    }

    // }}}
}
