// {{{ copyright

/********************************************************************
 *
 * $Id: ConfigWrapper.java,v 1.11 2000/06/06 17:00:02 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{ imports

import java.io.Serializable;

import java.util.Properties;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.ClassWrapper;

// }}}

// {{{ doc

/**
 * A ConfigWrapper implements the Configurable interface by wrapping itself
 * around a client Object, getting and setting its state when needed. <p>
 *
 * Standard ConfigWrappers can be constructed dynamically by passing
 * the object to wrap to the factory method {@link #makeWrapper makeWrapper}.
 * This will return a ConfigWrapper for the most derived class or baseclass
 * of the object, for which a wrapper is available, e.g.:
 *
 * <pre>
 * JFrame frame = new JFrame ();
 * ConfigWrapper wrapper = ConfigWrapper.makeWrapper(frame, "name");
 * </pre>
 *
 * will create a JFrameConfigWrapper. <p>
 *
 * You can make ConfigWrappers available in a package different from
 * <code>de.qfs.lib.config</code> by calling {@link #addPackage addPackage}.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.11 $
 */

// }}}
public class ConfigWrapper
    implements Configurable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (ConfigWrapper.class);

    /**
     * The name of the ConfigWrapper.
     */
    private String name;

    /**
     * An internal count for automatically generated names.
     */
    private static int count;

    /**
     * Dynamic creator for wrapper classes.
     */
    private static ClassWrapper wrapper =
        new ClassWrapper ("de.qfs.lib.config", "ConfigWrapper");

    // }}}

    // {{{ constructors

    /**
     * Create a new ConfigWrapper.
     */
    protected ConfigWrapper()
    {
        name = "object" + count++;
    }


    // }}}

    //----------------------------------------------------------------------
    // Static methods
    //----------------------------------------------------------------------
    // {{{ makeWrapper

    /**
     * A factory method that dynamically creates a ConfigWrapper for the
     * most derived class or baseclass of the client, for which a wrapper
     * is available.
     *
     * @param   client  The object to wrap.
     * @param   name    The name to give the new ConfigWrapper.
     *
     * @return  The new ConfigWrapper.
     */
    public static ConfigWrapper makeWrapper(Object client, String name)
    {
        ConfigWrapper cw = (ConfigWrapper) makeWrapper(client);
        cw.setConfigurableName(name);
        return cw;
    }

    /**
     * A factory method that dynamically creates a ConfigWrapper for the
     * most derived class or baseclass of the client, for which a wrapper
     * is available.
     *
     * @param   client  The object to wrap.
     *
     * @return  The new ConfigWrapper.
     */
    public static ConfigWrapper makeWrapper(Object client)
    {
        ConfigWrapper cw = (ConfigWrapper) wrapper.makeWrapper(client);
        if (cw == null) {
            cw = new ConfigWrapper ();
        }
        cw.setClient(client);
        return cw;
    }

    // }}}
    // {{{ addPackage

    /**
     * Add a package to be searched for ConfigWrappers. Packages added later
     * are considered more specific and are searched first.
     *
     * @param   packagename     The name of the package to add.
     */
    public static void addPackage(String packagename)
    {
        wrapper.addPackage(packagename);
    }

    // }}}

    //----------------------------------------------------------------------
    // Instance methods
    //----------------------------------------------------------------------
    // {{{ setClient

    /**
     * Set the client object for the ConfigWrapper.
     *
     * @param   client  The client object to set. Ignored.
     */
    protected void setClient(Object client)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // The Configurable interface
    //----------------------------------------------------------------------
    // {{{ setConfigurableName

    /**
     * Set the receiver's configuration name.
     *
     * @param   name    The name to set.
     */
    public void setConfigurableName (String name)
    {
        this.name = name;
    }

    // }}}
    // {{{ getConfigurableName

    /**
     * Get the receiver's configuration name.
     *
     * @return The receiver's configuration name.
     */
    public String getConfigurableName ()
    {
        return name;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     *
     * @return  An empty Properties object.
     */
    public Properties getConfigurableState()
    {
        return new Properties ();
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     *
     * @param   state   The state to set. Ignored.
     */
    public void setConfigurableState(Properties state)
    {
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     * This default implementation does nothing.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "registrationChanged(Configuration,String," +
                       "boolean,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "registered: " + registered);
        }
    }

    // }}}
}
