// {{{ copyright

/********************************************************************
 *
 * $Id: Configurable.java,v 1.9 2000/06/06 17:00:02 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

import java.io.Serializable;
import java.util.Properties;

/**
 * The Configurable interface is used together with the {@link Configuration
 * Configuration} class to make an Object's state persistent. <p>
 *
 * A Configurable has a name under which it is registered within the
 * Configuration. When the Configuration is saved, it queries all registered
 * Configurables for their state, which will be passed back to them, when the
 * Configuration is restored. <p>
 *
 * A Configurable should notify the Configuration via {@link
 * Configuration#stateChanged Configuration.stateChanged}, when its state
 * changes.
 *
 * @see Configuration
 * @author      Gregor Schmid
 * @version     $Revision: 1.9 $
 */
public interface Configurable
{
    /**
     * Set the Configurable's name.
     *
     * @param   name    The name to set.
     */
    public void setConfigurableName(String name);

    /**
     * Get the Configurable's name.
     *
     * @return The Configurable's name.
     */
    public String getConfigurableName();

    /**
     * Get the Configurable's state.
     *
     * @return The Configurable's state.
     */
    public Properties getConfigurableState();

    /**
     * Set the Configurable's state.
     *
     * @param   state   The state to set.
     */
    public void setConfigurableState(Properties state);

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept);
}
