// {{{ copyright

/********************************************************************
 *
 * $Id: JTableConfigWrapper.java,v 1.10 2000/06/06 17:00:02 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{  imports

import java.awt.Dimension;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.io.Serializable;

import java.util.Properties;

import javax.swing.SwingUtilities;
import javax.swing.JTable;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.TableColumnModelEvent;
import javax.swing.event.TableColumnModelListener;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;

import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A ConfigWrapper for swing.JTable objects.
 * The state saved includes the order of the columns as well as their
 * preferred witdth.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.10 $
 */
public class JTableConfigWrapper
    extends ConfigWrapper
    implements TableColumnModelListener, PropertyChangeListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (JTableConfigWrapper.class);

    /**
     * The wrapped table.
     */
    protected JTable table;

    /**
     * Helper for registration callbacks.
     */
    private RegistrationHelper regHelper = new RegistrationHelper ();

    // }}}

    // {{{ constructor

    /**
     * Create a new JTableConfigWrapper with the default name "Table".
     */
    public JTableConfigWrapper ()
    {
        setConfigurableName("Table");
    }

    // }}}

    // {{{ setClient

    /**
     * Set the client object for the ConfigWrapper.
     *
     * @param   client  The client object to set.
     */
    protected void setClient(Object client)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setClient(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "client: " + client +
                       ", this: " + this);
        }
        table = (JTable) client;
    }

    // }}}

    //----------------------------------------------------------------------
    // Configurable methods
    //----------------------------------------------------------------------
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "this: " + this);
        }
        Properties state = new Properties ();

        TableColumnModel cm = table.getColumnModel();
        int num = cm.getColumnCount();
        for (int i = 0; i < cm.getColumnCount(); i++) {
            TableColumn col = cm.getColumn(i);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "getConfigurableState()",
                           "Column " + i + "- id: " +
                           col.getIdentifier().toString() +
                           ", width: " + col.getPreferredWidth());
            }
            state.put("id." + i, col.getIdentifier().toString());
            state.put("width." + i, "" + col.getPreferredWidth());
        }
        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "state: " + state);
        }

        Dimension size = table.getSize();
        table.setSize(0, 0);
        TableColumnModel cm = table.getColumnModel();

        for (int i = 0; ; i++) {
            String id = state.getProperty("id." + i);
            String width = state.getProperty("width." + i);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "setConfigurableState(Properties)",
                           "id." + i + ": " + id +
                           ", width." + i + ": " + width);
            }
            if (id == null || width == null) {
                break;
            }
            try {
                int w = Integer.parseInt(width);
                for (int j = 0; j < cm.getColumnCount(); j++) {
                    if (id.equals(cm.getColumn(j).getIdentifier()
                                  .toString())) {
                        cm.moveColumn(j, i);
                        cm.getColumn(i).setPreferredWidth(w);
                        break;
                    }
                }
            } catch (NumberFormatException ex) {
                logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
                break;
            }
        }

        table.setSize(size);
        table.getTableHeader().resizeAndRepaint();
        table.repaint();
    }

    // }}}

    // {{{ getState

    /**
     * Get the state of the Table.
     *
     * @return  An Object array with two elements, an array of the ids
     *          of the columns and an array ot their preferred widths.
     */
    public Serializable getState ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getState()", "this: " + this);
        }

        TableColumnModel cm = table.getColumnModel();
        int num = cm.getColumnCount();
        Object[] ids = new Object[num];
        int[] widths = new int[num];
        for (int i = 0; i < cm.getColumnCount(); i++) {
            TableColumn col = cm.getColumn(i);
            ids[i] = col.getIdentifier();
            widths[i] = col.getPreferredWidth();
            if (logger.level >= Log.DBGDETAIL) {
                logger.log(Log.DBGDETAIL, "getState()",
                           "Column " + i + ": " + ids[i] + ", " + widths[i]);
            }
        }
        Object[] state = {ids, widths};
        return (Serializable) ((Object) state);
    }

    // }}}
    // {{{ setState

    /**
     * Set the state of the Table.
     *
     * @param   state   An Object array that should be of the form returned
     *                  by {@link #getState getState}.
     */
    public void setState (Object state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setState(Object)",
                        logger.level < Log.MTDDETAIL ? "" :
                        "state: " + state);
        }

        try {
            final Object[] ids = (Object[]) ((Object[])state)[0];
            final int[] widths = (int[]) ((Object[])state)[1];
            SwingUtil.invokeAndWait(new Runnable() {
                public void run () {
                    restoreColumns (ids, widths);
                }
            });
        } catch (ClassCastException ex) {
            if (logger.level >= Log.ERR) {
                logger.log ("setState(Object)", ex);
            }
        }
    }

    // }}}

    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "registrationChanged(Configuration,String," +
                       "boolean,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "registered: " + registered);
        }
        int ret = regHelper.registrationChanged(config, registered);
        if (table != null && ret != 0) {
            TableColumnModel tcm = table.getColumnModel();
            if (ret == 1) {
                tcm.addColumnModelListener(this);
                for (int i = 0; i < tcm.getColumnCount(); i++) {
                    tcm.getColumn(i).addPropertyChangeListener(this);
                }
            } else if (ret == -1) {
                tcm.removeColumnModelListener(this);
                for (int i = 0; i < tcm.getColumnCount(); i++) {
                    tcm.getColumn(i).removePropertyChangeListener(this);
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // TableColumnModelListener interface
    //----------------------------------------------------------------------
    // {{{ columnAdded

    public void columnAdded(TableColumnModelEvent e)
    {
        regHelper.stateChanged(this);
    }

    // }}}
    // {{{ columnRemoved

    public void columnRemoved(TableColumnModelEvent e)
    {
        regHelper.stateChanged(this);
    }

    // }}}
    // {{{ columnMoved

    public void columnMoved(TableColumnModelEvent e)
    {
        regHelper.stateChanged(this);
    }

    // }}}
    // {{{ columnMarginChanged

    public void columnMarginChanged(ChangeEvent e)
    {
    }

    // }}}
    // {{{ columnSelectionChanged

    public void columnSelectionChanged(ListSelectionEvent e)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // PropertyChangeListener interface
    //----------------------------------------------------------------------
    // {{{ propertyChanged

    public void propertyChange(PropertyChangeEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "propertyChanged(PropertyChangeEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        if (e.getPropertyName().equals
            (TableColumn.COLUMN_WIDTH_PROPERTY)) {
            regHelper.stateChanged(this);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ restoreColumns

    private void restoreColumns (Object[] ids, int[] widths)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restoreColumns(Object[],int[])",
                        logger.level < Log.MTDDETAIL ? "" :
                        "ids: " + ids + ", " +
                        "widths: " + widths);
        }

        Dimension size = table.getSize();
        table.setSize(0, 0);
        TableColumnModel cm = table.getColumnModel();
        // Then reorder them and set their old witdh
        for (int i = 0; i < ids.length; i++) {
            if (logger.level >= Log.DBGDETAIL) {
                logger.log(Log.DBGDETAIL, "restoreColumns(Object[],int[])",
                           "Column " + i + ": " + ids[i] + ", " + widths[i]);
            }
            try {
                cm.moveColumn(cm.getColumnIndex(ids[i]), i);
                cm.getColumn(i).setPreferredWidth(widths[i]);
            } catch (IllegalArgumentException ex) {
                // identifier no longer valid - ignore
            }
        }
        table.setSize(size);
        table.getTableHeader().resizeAndRepaint();
        table.repaint();
    }

    // }}}
}
