// {{{ copyright

/********************************************************************
 *
 * $Id: JToolBarConfigWrapper.java,v 1.7 2001/01/19 19:11:55 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{ imports

import java.awt.Point;

import java.io.Serializable;

import java.util.Properties;

import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.plaf.basic.BasicToolBarUI;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This ConfigWrapper saves and restores the State of a JToolBar. The saved
 * state contains information about whether the toolbar is floating or docked
 * and about its position, either inside its floating Frame or its parent. <p>
 *
 * TODO: Listen for changes and call {@link Configuration#stateChanged
 * Configuration.stateChanged}.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 */
public class JToolBarConfigWrapper extends ConfigWrapper
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (JToolBarConfigWrapper.class);

    /**
     * The JToolBar to wrap.
     */
    protected JToolBar toolbar;

    // }}}

    // {{{ constructor

    /**
     * Create a new JToolBarConfigWrapper with the default name 'toolbar'.
     */
    public JToolBarConfigWrapper ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "JToolBarConfigWrapper()", "");
        }
        setConfigurableName("toolbar");
    }

    // }}}

    // {{{ setClient

    /**
     * Set the client object for the ConfigWrapper.
     *
     * @param   client  The client object to set.
     */
    protected void setClient(Object client)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setClient(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "client: " + client);
        }
        toolbar = (JToolBar)client;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "");
        }
        Properties state = new Properties ();
        BasicToolBarUI ui = (BasicToolBarUI) toolbar.getUI();

        state.put("floating", "" + ui.isFloating());
        if (ui.isFloating()) {
            Point pos = SwingUtilities.getRootPane(toolbar).getParent()
                .getLocation();
            state.put("x", "" + pos.x);
            state.put("y", "" + pos.y);
        } else {
            String pos;
            if (toolbar.getParent() == null) {
                // no guess possible
                pos = "North";
            } else {
                int x = toolbar.getLocation().x -
                    toolbar.getParent().getInsets().left;
                int y = toolbar.getLocation().y -
                    toolbar.getParent().getInsets().top;
                int width = toolbar.getSize().width +
                    toolbar.getParent().getInsets().left +
                    toolbar.getParent().getInsets().right;
                // Try to guess the position
                if (width == toolbar.getParent().getSize().width) {
                    if (y == 0) {
                        pos = "North";
                    } else {
                        pos = "South";
                    }
                } else {
                    if (x == 0) {
                        pos = "West";
                    } else {
                        pos = "East";
                    }
                }
            }
            state.put("anchor", pos);
        }

        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "state: " + state);
        }

        String floating = state.getProperty("floating");
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setConfigurableState(Properties)",
                       "floating: " + floating);
        }
        if (floating == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setConfigurableState(Properties)",
                           "Invalid state: " +
                           "floating: " + floating);
            }
            return;
        }

        BasicToolBarUI ui = (BasicToolBarUI) toolbar.getUI();
        if (Boolean.valueOf(floating).booleanValue()) {
            String x = state.getProperty("x");
            String y = state.getProperty("y");
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "setConfigurableState(Properties)",
                           "x: " + x + ", y: " + y);
            }
            if (x == null || y == null) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "setConfigurableState(Properties)",
                               "Invalid state: " +
                               "x: " + x + ", y: " + y);
                }
                return;
            }
            try {
                Point pos = new Point (Integer.parseInt(x),
                                       Integer.parseInt(y));
                // Should probably suffice, but doesn't
                ui.setFloating(true, pos);
                // So we add...
                SwingUtilities.getRootPane(toolbar).getParent()
                    .setLocation(pos);
            } catch (NumberFormatException ex) {
                logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
            }
        } else {
            String anchor = state.getProperty("anchor");
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "setConfigurableState(Properties)",
                           "anchor: " + anchor);
            }
            ui.setFloating(false, null);
            toolbar.getParent().add(toolbar, anchor);
            if ("East".equals(anchor) || "West".equals(anchor)) {
                ui.setOrientation(toolbar.VERTICAL);
            } else {
                ui.setOrientation(toolbar.HORIZONTAL);
            }
            toolbar.revalidate();
        }
    }

    // }}}
}
