// {{{ copyright

/********************************************************************
 *
 * $Id: EventQueueMark.java,v 1.3 2001/02/04 12:19:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.SwingUtilities;
import javax.swing.Timer;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * This little tool is useful to distinguish between artificial and
 * user-generated events. Often it is necessary to decide, whether an event,
 * e.g. a change in some component's selection, was generated by the
 * application itself in responce to a certain action (in which case it should
 * be ignored), or was caused explicitely by the user (so some response is
 * required). <p>
 *
 * To determine whether an event is artificial, create an EventQueueMark and
 * {@link #activate() activate} it right before and after the action that
 * causes the event - before to ignore artificial events that are passed to
 * listeners directly, bypassing the event queue, and after to ignore events
 * collected on the queue during the action. In the corresponding listener,
 * use {@link #isActive isActive} to determine whether the mark is still
 * active on the event queue, meaning the event can be ignored. <p>
 *
 * If the artificial event is not generated immediately, but after some delay,
 * use {@link #activate(int) activate(int delay)} with a similar, possibly
 * slightly longer delay.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.3 $
 * @since       0.98.0
 */

// }}}
public class EventQueueMark
    implements Runnable, ActionListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (EventQueueMark.class);

    /**
     * Flag to determine whether the mark is active.
     */
    protected boolean active;

    /**
     * That many incantations need to be cleared.
     */
    protected int toClear;

    /**
     * That many clearances need to be skipped.
     */
    protected int toSkip;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ EventQueueMark

    /**
     * Create a new EventQueueMark.
     */
    public EventQueueMark ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "EventQueueMark()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ activate()

    /**
     * Activate the mark until the the events currently on the event queue
     * are removed.
     */
    public synchronized void activate()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "activate()", "");
        }
        toClear++;
        active = true;
        SwingUtilities.invokeLater(this);
    }

    // }}}
    // {{{ activate(int)

    /**
     * Activate the mark until some time has passed.
     *
     * @param   delay   The time to wait in ms.
     */
    public synchronized void activate(int delay)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "activate(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "delay: " + delay);
        }
        if (delay <= 0) {
            activate();
            return;
        }

        Timer timer = new Timer (delay, this);
        timer.setRepeats(false);
        toClear++;
        active = true;
        timer.start();
    }

    // }}}
    // {{{ deactivate

    /**
     * Deactivate the mark immediately.
     */
    public synchronized void deactivate()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "deactivate()", "");
        }
        toSkip = toClear;
        toClear = 0;
        active = false;
    }

    // }}}
    // {{{ isActive

    /**
     * Check whether the mark is active.
     *
     * @return  True if the mark is active.
     */
    public synchronized boolean isActive()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isActive()",
                       "active: " + active);
        }
        return active;
    }

    // }}}

    //----------------------------------------------------------------------
    // The ActionListener interface
    //----------------------------------------------------------------------
    // {{{ actionPerformed

    /**
     * Called when a timer fires. Decrement active count and deactivate the
     * mark if it reaches 0.
     *
     * @param   e       Ignored.
     */
    public void actionPerformed(ActionEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "actionPerformed(ActionEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        run();
    }

    // }}}

    //----------------------------------------------------------------------
    // The Runnable interface
    //----------------------------------------------------------------------
    // {{{ run

    /**
     * Zun method for {@link SwingUtilities#invokeLater
     * SwingUtilities.invokeLater}. Decrement active count and deactivate the
     * mark if it reaches 0.
     */
    public synchronized void run()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "run()",
                       "active: " + active +
                       ", toClear:" + toClear +
                       ", toSkip: " + toSkip);
        }
        if (toSkip > 0) {
            toSkip--;
        } else if (toClear > 0) {
            toClear--;
            if (toClear == 0) {
                active = false;
            }
        } else {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "run()",
                           "Unexpected invocation");
            }
            active = false;
        }
    }

    // }}}
}
