// {{{ copyright

/********************************************************************
 *
 * $Id: MultiPane.java,v 1.12 2000/11/14 11:29:30 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.SwingUtilities;

import de.qfs.lib.config.Configurable;
import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.JSplitPaneConfigWrapper;
import de.qfs.lib.config.RegistrationHelper;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A MultiPane is very similar to a JSplitPane and is implemented with the
 * help of one. <p>
 *
 * It adds to the JSplitPane API methods for displaying either one of the
 * two components or both. <p>
 *
 * MultiPane implements the {@link Configurable Configurable} interface. The
 * state saved is the location of the JSplitPane divider and the showing state,
 * i.e. whether one or both of the components are shown.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.12 $
 */
public class MultiPane
    extends JPanel
    implements Configurable, PropertyChangeListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (MultiPane.class);

    /**
     * Indicates that both components are shown.
     */
    public final static int SHOW_BOTH = 0;

    /**
     * Indicates that only the first component is shown.
     */
    public final static int SHOW_FIRST = 1;

    /**
     * Indicates that only the second component is shown.
     */
    public final static int SHOW_SECOND = 2;

    /**
     * Use a horizontal layout for the components.
     */
    public final static int HORIZONTAL = JSplitPane.HORIZONTAL_SPLIT;

    /**
     * Use a vertical layout for the components.
     */
    public final static int VERTICAL = JSplitPane.VERTICAL_SPLIT;

    /**
     * The top/left component.
     */
    private Component first;

    /**
     * The bottom/right component.
     */
    private Component second;

    /**
     * The split pane used internally.
     */
    private JSplitPane split;

    /**
     * The current showing state.
     */
    private int shown = SHOW_BOTH;

    /**
     * The target location of the split pane divider.
     */
    private int divider = -1;

    /**
     * The relative target location of the split pane divider.
     */
    private double relDivider = -1;

    /**
     * Flag to avoid multiple override of divider settings.
     */
    private boolean override = true;

    /**
     * ConfigWrapper to help with the split pane layout.
     */
    private JSplitPaneConfigWrapper wrapper;

    /**
     * The configurable name of the MultiPane.
     */
    private String name;

    /**
     * The registered event listeners.
     */
    private Vector listeners = new Vector ();

    /**
     * Helper for registration callbacks.
     */
    private RegistrationHelper regHelper = new RegistrationHelper ();

    // }}}

    // {{{ constructors

    /**
     * Create a new MultiPane with a name of "MultiPane", a default divider
     * location of 0.5 and a default showing state of {@link #SHOW_BOTH
     * SHOW_BOTH}.
     *
     * @param   orientation     The orientation, one of
     *                          {@link #HORIZONTAL HORIZONTAL} or
     *                          {@link #VERTICAL VERTICAL}
     *
     * @throws  IllegalArgumentException If the show parameter is not one of
     * {@link #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or
     * {@link #SHOW_SECOND SHOW_SECOND}.
     */
    public MultiPane(int orientation)
    {
        this(orientation, null, null);
    }

    /**
     * Create a new MultiPane with a name of "MultiPane", a default divider
     * location of 0.5 and a default showing state of {@link #SHOW_BOTH
     * SHOW_BOTH}.
     *
     * @param   orientation     The orientation, one of
     *                          {@link #HORIZONTAL HORIZONTAL} or
     *                          {@link #VERTICAL VERTICAL}
     * @param   firstComponent  The top/left Component.
     * @param   secondComponent The bottom/right Component.
     *
     * @throws  IllegalArgumentException If the show parameter is not one of
     * {@link #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or
     * {@link #SHOW_SECOND SHOW_SECOND}.
     */
    public MultiPane(int orientation, Component firstComponent,
                     Component secondComponent)
    {
        this("MultiPane", orientation, firstComponent, secondComponent,
             0.5, SHOW_BOTH);
    }

    /**
     * Create a new MultiPane.
     *
     * @param   name            The {@link Configurable Configurable} name.
     * @param   orientation     The orientation, one of
     *                          {@link #HORIZONTAL HORIZONTAL} or
     *                          {@link #VERTICAL VERTICAL}
     * @param   firstComponent  The top/left Component.
     * @param   secondComponent The bottom/right Component.
     * @param   divider         The default divider location.
     * @param   show            The initial display as in {@link #setShowing
     *                          setShowing}.
     *
     * @throws  IllegalArgumentException If the show parameter is not one of
     * {@link #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or
     * {@link #SHOW_SECOND SHOW_SECOND}.
     */
    public MultiPane(String name, int orientation,
                     Component firstComponent,
                     Component secondComponent,
                     int divider, int show)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "MultiPane(String,int,Component,Component,int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "orientation: " + orientation + ", " +
                       "firstComponent: " + firstComponent + ", " +
                       "secondComponent: " + secondComponent + ", " +
                       "divider: " + divider + ", " +
                       "show: " + show);
        }
        this.name = name;
        setLayout(new BorderLayout ());
        setBorder(null);
        first = firstComponent;
        second = secondComponent;
        split = new JSplitPane (orientation);
        split.addPropertyChangeListener(this);
        split.setBorder(null);
        wrapper = new JSplitPaneConfigWrapper (split, name, divider);
        this.divider = divider;
        shown = -1;
        setShowing(show);
    }

    /**
     * Create a new MultiPane.
     *
     * @param   name            The {@link Configurable Configurable} name.
     * @param   orientation     The orientation, one of
     *                          {@link #HORIZONTAL HORIZONTAL} or
     *                          {@link #VERTICAL VERTICAL}
     * @param   firstComponent  The top/left Component.
     * @param   secondComponent The bottom/right Component.
     * @param   divider         The default divider location.
     * @param   show            The initial display as in {@link #setShowing
     *                          setShowing}.
     *
     * @throws  IllegalArgumentException If the show parameter is not one of
     * {@link #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or
     * {@link #SHOW_SECOND SHOW_SECOND}.
     */
    public MultiPane(String name, int orientation,
                     Component firstComponent,
                     Component secondComponent,
                     double divider, int show)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "MultiPane(String,int,Component,Component,double,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "orientation: " + orientation + ", " +
                       "firstComponent: " + firstComponent + ", " +
                       "secondComponent: " + secondComponent + ", " +
                       "divider: " + divider + ", " +
                       "show: " + show);
        }
        this.name = name;
        setLayout(new BorderLayout ());
        setBorder(null);
        first = firstComponent;
        second = secondComponent;
        split = new JSplitPane (orientation);
        split.addPropertyChangeListener(this);
        split.setBorder(null);
        wrapper = new JSplitPaneConfigWrapper (split, name, divider);
        this.relDivider = divider;
        shown = -1;
        setShowing(show);
    }

    // }}}

    // {{{ getSplitPane

    /**
     * Get the JSplitPane used internally by the MultiPane.
     *
     * @return  The internal JSplitPane.
     */
    public JSplitPane getSplitPane()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getSplitPane()",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name);
        }
        return split;
    }

    // }}}
    // {{{ getShowing

    /**
     * Get the current showing state of the MultiPane, one of {@link
     * #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or {@link
     * #SHOW_SECOND SHOW_SECOND}.
     *
     * @return  The current showing state.
     */
    public final int getShowing()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getShowing()",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name);
        }
        return shown;
    }

    // }}}
    // {{{ setShowing

    /**
     * Set which component(s) to display. Must be one of {@link #SHOW_BOTH
     * SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or {@link #SHOW_SECOND
     * SHOW_SECOND}.
     *
     * @param   show    The new showing state.
     *
     * @throws  IllegalArgumentException If the show parameter is not one of
     * {@link #SHOW_BOTH SHOW_BOTH}, {@link #SHOW_FIRST SHOW_FIRST} or
     * {@link #SHOW_SECOND SHOW_SECOND}.
     */
    public void setShowing(int show)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setShowing(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name +
                       ", show: " + show +
                       " (divider: " + divider + ")");
        }
        if (show < SHOW_BOTH && show > SHOW_SECOND) {
            throw new IllegalArgumentException
                ("setShowing: argument " + show + " must be oned of " +
                 "SHOW_BOTH, SHOW_FIRST or SHOW_SECOND");
        }

        if (shown == show) {
            if (shown == SHOW_BOTH && divider >= 0) {
                final int tmp = divider;
                SwingUtilities.invokeLater(new Runnable () {
                    public void run() {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "run()",
                                       "Setting tmp:" + tmp +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(tmp);
                    }
                });
            }
            return;
        }

        if (shown == SHOW_BOTH) {
            // save current divider position
            //  divider = split.getDividerLocation();
        }

        java.awt.Component focusOwner = SwingUtilities.findFocusOwner(this);
        if (logger.level >= Log.DBGDETAIL) {
            logger.log(Log.DBGDETAIL, "setShowing(int)",
                       "focusOwner: " + focusOwner);
        }
        if (focusOwner != null) {
            requestFocus();
        }
        shown = show;
        switch (shown) {
        case SHOW_BOTH:
            final int tmp = divider;
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "setShowing(int)",
                           "Old divider: " + tmp);
            }
            split.setLeftComponent(first);
            split.setRightComponent(second);
            add(split, BorderLayout.CENTER);
            // use invoke later to set the divider location, since changing
            // the components can add a pending property change for the
            // divider that has to be overridden.
            if (tmp >= 0) {
                SwingUtilities.invokeLater(new Runnable () {
                    public void run() {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "run()",
                                       "Setting tmp:" + tmp +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(tmp);
                    }
                });
            } else if (relDivider >= 0) {
                SwingUtilities.invokeLater(new Runnable () {
                    public void run() {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "run()",
                                       "Setting relDivider: " + relDivider +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(relDivider);
                    }
                });
            }
            if (focusOwner != null) {
                focusOwner.requestFocus();
            }
            break;
        case SHOW_FIRST:
            removeAll();
            add(first, BorderLayout.CENTER);
            if (focusOwner != null) {
                if (SwingUtilities.isDescendingFrom(focusOwner, first)) {
                    focusOwner.requestFocus();
                } else {
                    first.requestFocus();
                }
            }
            break;
        case SHOW_SECOND:
            removeAll();
            add(second, BorderLayout.CENTER);
            if (focusOwner != null) {
                if (SwingUtilities.isDescendingFrom(focusOwner, second)) {
                    focusOwner.requestFocus();
                } else {
                    second.requestFocus();
                }
            }
            break;
        }
        revalidate();
        if (focusOwner != null) {
            //  focusOwner.requestFocus();
        }
        fireShowingChanged(shown);
    }

    // }}}
    // {{{ getFirstComponent

    /**
     * Get the top/left component of the MultiPane.
     *
     * @return  The top/left component.
     */
    public final Component getFirstComponent()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getFirstComponent()",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name);
        }
        return first;
    }

    // }}}
    // {{{ setFirstComponent

    /**
     * Set the top/left component of the MultiPane.
     *
     * @param   first   The new top/left component.
     */
    public final void setFirstComponent(Component first)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setFirstComponent(Component)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name +
                       ", first: " + first);
        }
        this.first = first;
        if (shown == SHOW_BOTH) {
            split.setLeftComponent(first);
        }
    }

    // }}}
    // {{{ getSecondComponent

    /**
     * Get the bottom/right component of the MultiPane.
     *
     * @return  The bottom/right component.
     */
    public final Component getSecondComponent()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getSecondComponent()",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name);
        }
        return second;
    }

    // }}}
    // {{{ setSecondComponent

    /**
     * Set the bottom/right component of the MultiPane.
     *
     * @param   second  The new bottom/right component.
     */
    public final void setSecondComponent(Component second)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSecondComponent(Component)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name +
                       ", second: " + second);
        }
        this.second = second;
        if (shown == SHOW_BOTH) {
            split.setLeftComponent(second);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Events
    //----------------------------------------------------------------------
    // {{{ addMultiPaneListener

    /**
     * Add a MultiPaneListener to the MultiPane.
     *
     * @param   listener        The MultiPaneListener to add.
     */
    public void addMultiPaneListener(MultiPaneListener listener)
    {
        listeners.addElement(listener);
    }

    // }}}
    // {{{ removeMultiPaneListener

    /**
     * Remove a MultiPaneListener from the MultiPane.
     *
     * @param   listener        The MultiPaneListener to remove.
     */
    public void removeMultiPaneListener(MultiPaneListener listener)
    {
        listeners.removeElement(listener);
    }

    // }}}
    // {{{ fireShowingChanged

    /**
     * Notify all MultiPaneListeners of a change in the showing state.
     *
     * @param   showing The new showing state.
     */
    private void fireShowingChanged(int showing)
    {
        Vector tmp = (Vector) listeners.clone();
        MultiPaneEvent event = new MultiPaneEvent (this, showing);

        for (Enumeration e = tmp.elements(); e.hasMoreElements();) {
            ((MultiPaneListener) e.nextElement()).showingChanged(event);
        }
        regHelper.stateChanged(this);
    }

    // }}}

    //----------------------------------------------------------------------
    // The Configurable interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableName

    /**
     * Get the receiver's configuration name.
     *
     * @return The receiver's configuration name.
     */
    public final String getConfigurableName()
    {
        return name;
    }

    // }}}
    // {{{ setConfigurableName

    /**
     * Set the receiver's configuration name.
     *
     * @param   name    The name to set.
     */
    public final void setConfigurableName(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableName(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        this.name = name;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state. It is composed of the location of the
     * JSplitPane divider and the showing state.
     *
     * @return  The state of the MultiPane.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "");
        }
        Properties state;
        if (shown == SHOW_BOTH) {
            state = wrapper.getConfigurableState();
        } else {
            state = new Properties ();
            if (divider >= 0) {
                state.put("divider", "" + divider);
            } else if (relDivider >= 0) {
                state.put("divider", "" + relDivider);
            }
        }
        state.put("show", "" + shown);
        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + name +
                       ", state: " + state);
        }
        wrapper.setConfigurableState(state);

        String pos = state.getProperty("divider");
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setConfigurableState(Properties)",
                       "divider: " + pos);
        }
        if (pos == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setConfigurableState(Properties)",
                           "Invalid state: " +
                           "divider: " + pos);
            }
            return;
        }

        try {
            divider = Integer.parseInt(pos);
        } catch (NumberFormatException ex) {
            try {
                relDivider = new Double (pos).doubleValue();
                divider = -1;
            } catch (NumberFormatException ex2) {
                logger.log(Log.WRN, "setConfigurableState(Properties)", ex2);
                return;
            }
        }

        int tmp = divider;

        String show = state.getProperty("show");
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setConfigurableState(Properties)",
                       "show: " + show);
        }
        if (show == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setConfigurableState(Properties)",
                           "Invalid state: " +
                           "show: " + show);
            }
            return;
        }
        try {
            final int s = Integer.parseInt(show);
            SwingUtil.invokeAndWait(new Runnable() {
                public void run () {
                    setShowing(s);
                }
            });
        } catch (NumberFormatException ex) {
            logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
            return;
        } catch (IllegalArgumentException ex) {
            logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
            return;
        }

        divider = tmp;
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "registrationChanged(Configuration,String," +
                       "boolean,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "registered: " + registered);
        }
        regHelper.registrationChanged(config, registered);
    }

    // }}}

    //----------------------------------------------------------------------
    // The PropertyChangeListener interface
    //----------------------------------------------------------------------
    // {{{ propertyChange

    /**
     * Notify the MultiPane of a change in the JSplitPane.
     *
     * @param   e       The event containing the details of the change.
     */
    public void propertyChange(PropertyChangeEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "propertyChanged(PropertyChangeEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "myName: " + getConfigurableName() +
                       ", propertyName: " + e.getPropertyName());
        }
        if (e.getPropertyName().equals
               (JSplitPane.LAST_DIVIDER_LOCATION_PROPERTY )) {
            if (! isShowing() &&
                (divider >= 0 || relDivider >= 0)) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "propertyChange(PropertyChangeEvent)",
                               "not showing, ignored");
                }
                if (override) {
                    if (divider >= 0) {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG,
                                       "propertyChange(PropertyChangeEvent)",
                                       "Setting divider: " + divider +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(divider);
                    } else {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG,
                                       "propertyChange(PropertyChangeEvent)",
                                       "Setting divider: " + divider +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(divider);
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG,
                                       "propertyChange(PropertyChangeEvent)",
                                       "Setting relDivider: " + relDivider +
                                       ", myName: " + name);
                        }
                        split.setDividerLocation(relDivider);
                    }
                    override = false;
                }
            } else {
                divider = split.getDividerLocation();
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "propertyChange(PropertyChangeEvent)",
                               "divider: " + divider);
                }
                regHelper.stateChanged(this);
            }
        }
    }

    // }}}
}
