// {{{ copyright

/********************************************************************
 *
 * $Id: SwingUtil.jpp,v 1.28 2001/01/19 19:12:09 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.lang.reflect.InvocationTargetException;


import java.awt.ActiveEvent;


import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Label;
import java.awt.MenuComponent;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.PaintEvent;
import java.awt.event.WindowEvent;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


import javax.swing.Action;
import javax.swing.KeyStroke;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JPopupMenu;
import javax.swing.JPasswordField;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.JViewport;
import javax.swing.Scrollable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.text.JTextComponent;
import javax.swing.text.Keymap;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeModel;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A collection of utilities and Swing problem workarounds that go beyond
 * {@link javax.swing.SwingUtilities SwingUtilities}.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.28 $
 */
public abstract class SwingUtil
{
    // {{{ variables

    /**
     * Identifier for no events.
     */
    public final static int NO_EVENTS = 0;

    /**
     * Identifier for all events.
     */
    public final static int ALL_EVENTS = ~0;

    /**
     * Identifier for paint events.
     */
    public final static int PAINT_EVENTS = 1;

    /**
     * Identifier for input events.
     */
    public final static int INPUT_EVENTS = 2;

    /**
     * Identifier for the rest of the events. Must include all unused types
     * for compatibility with possible future extensions.
     */
    public final static int OTHER_EVENTS = ~3;

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (SwingUtil.class);

    /**
     * Keymap with the original binding for enter.
     */
    private static Keymap defaultKeymap;

    /**
     * ID counter for DoneEvent ids.
     */
    private static int doneID;

    // }}}

    //----------------------------------------------------------------------
    // EventLoop
    //----------------------------------------------------------------------
    // {{{ invokeAndWait

  /**
   * Invoke a thread inside the swing EventLoop and wait for it to finish.
   *
   * @param     runnable The Runnable to invoke.
   *
   * @return    True if everything was OK, false, if the current Thread was
   *            interrupted while waititing for the EventLoop.
   */
    public static boolean invokeAndWait (Runnable runnable)
    {
        if (SwingUtilities.isEventDispatchThread()) {
            runnable.run();
        } else {
            try {
                SwingUtilities.invokeAndWait(runnable);
            } catch (InvocationTargetException ex) {
                if (ex.getTargetException() instanceof RuntimeException) {
                    throw (RuntimeException) ex.getTargetException();
                } else {
                    // impossible ?!?
                    ex.getTargetException().printStackTrace();
                }
            } catch (InterruptedException ex) {
                return false;
            }
        }
        return true;
    }

    // }}}
    // {{{ interface UnsafeRunnable

    /**
     * This interface is similar to Runnable, except that the run method
     * may return a value and throw any kind of Exception.
     */
    public static interface UnsafeRunnable
    {
        /**
         * Execute the unsafe runnable.
         *
         * @return      Whatever the UnsafeRunnable decides to return.
         *
         * @throws      Exception       If anything goes wrong.
         */
        public Object run() throws Exception;
    }

    // }}}
    // {{{ withEvents

    /**
     * Execute some code in a background Thread, while continuing to handle
     * AWT Events from the current Thread. InputEvents like mouse or keyboard
     * input can optionally be discarded (useful together with a wait cursor).
     * This works only if called from an AWT EventLoop Thread. When called from
     * some other Thread, the Runnable simply gets executed directly. In this
     * case there is no way to block input events.
     *
     * @param   runnable        The UnsafeRunnable to execute.
     * @param   noInput         If true, InputEvents will be discarded.
     *
     * @return  The result from runnable's execution.
     *
     * @throws  RuntimeException        If runnable's run method
     *          throws a RuntimeException, it is ismply passed on
     * @throws  InvocationTargetException       If runnable's run method*
     *          throws an Exception.
     *
     * @deprecated      Please use
     *                  {@link #withEvents(SwingUtil.UnsafeRunnable,int,int)
     *                  withEvents} instead.
     */
    public static Object withEvents(final UnsafeRunnable runnable,
                                    final boolean noInput)
        throws InvocationTargetException
    {
        if (noInput) {
            return withEvents(runnable, PAINT_EVENTS | OTHER_EVENTS,
                              INPUT_EVENTS);
        } else {
            return withEvents(runnable, ALL_EVENTS, NO_EVENTS);
        }
    }

    /**
     * Execute some code in a background thread, while continuing to handle
     * AWT events from the current tread. <p>
     *
     * This method distinguishes between three kinds of events identified by
     * <dl>
     * <dt>PAINT_EVENTS</dt><dd>All kinds of paint events.</dd>
     * <dt>INPUT_EVENTS</dt><dd>All kind of input events, e.g. keyboard or
     * mouse events.</dd>
     * <dt>OTHER_EVENTS</dt><dd>All the rest.</dd>
     * </dl>
     * For each of these kinds of events, there are three possible ways to
     * handle them:
     * <dl>
     * <dt>dispatch</dt><dd>Dispatch the event immediately as it arrives. This
     * is the default for all events.</dd>
     * <dt>delay</dt><dd>Delay the event until the background thread is
     * finished. Useful if handling the event could interfere with the
     * background thread.</dd>
     * <dt>discard</dt>
     * <dd> Discard the event as if it never happended. Makes sense for input
     * events if a wait cursor is being displayed while the background thread
     * is active.</dd>
     * </dl>
     *
     * This works only if called from an AWT EventLoop Thread. When called from
     * some other Thread, the Runnable simply gets executed directly. In this
     * case there is no way to block or delay events. <p>
     *
     * This method is more or less reentrant, meaning that recusive calls don't
     * fail horribly, but they can lead to confusion since the last call will
     * determine the delayed and discarded event types for all calls.
     *
     * @param   runnable        The UnsafeRunnable to execute.
     * @param   delay           Binary OR of event types that should be
     *                          delayed.
     * @param   discard         Binary OR of event types that should be
     *                          discarded.
     *
     * @return  The result from runnable's execution.
     *
     * @throws  RuntimeException        If runnable's run method
     *          throws a RuntimeException, it is ismply passed on
     * @throws  InvocationTargetException       If runnable's run method*
     *          throws an Exception.
     */
    public static Object withEvents(final UnsafeRunnable runnable,
                                    int delay, int discard)
        throws InvocationTargetException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "withEvents(UnsafeRunnable,int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "runnable: " + runnable + ", " +
                       "delay: " + delay + ", " +
                       "discard: " + discard);
        }
        if (! SwingUtilities.isEventDispatchThread()) {
                // cannot block events!
            try {
                return runnable.run();
            } catch (RuntimeException ex) {
                throw ex;
            } catch (Exception ex) {
                return new InvocationTargetException(ex);
            }
        }

        // I hate this hack...
        final Exception[] exHolder = new Exception[1];
        final Object[] oHolder = new Object[1];
        final int id = doneID++;
        new Thread () {
            public void run()
            {
                try {
                    oHolder[0] = runnable.run();
                } catch (Exception ex) {
                    exHolder[0] = ex;
                }

                Toolkit.getDefaultToolkit().getSystemEventQueue()
                    .postEvent(new DoneEvent(id));
            }
        }.start();
        poll(delay, discard, id);

        if (exHolder[0] != null) {
            if (exHolder[0] instanceof RuntimeException) {
                throw (RuntimeException) exHolder[0];
            }
            throw new InvocationTargetException(exHolder[0]);
        }
        return oHolder[0];
    }

    // }}}
    // {{{ poll

    /**
     * Helper method for polling and dispatching AWT events.
     */
    private static void poll(int delay, int discard, int id)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "poll(int,int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "delay: " + delay + ", " +
                       "discard: " + discard + ", " +
                       "id: " + id);
        }
        java.awt.EventQueue queue = EventQueue.instance();
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "poll(int,int,int)",
                       "queue: " + queue);
        }

        boolean done = false;
        List eventStack = new ArrayList ();
        while (!done) {
            try {
                AWTEvent event = queue.getNextEvent();
                AWTEvent orig = event;
                if (event instanceof EventQueue.WrapperEvent) {
                    event = ((EventQueue.WrapperEvent) event)
                        .getWrappedEvent();
                }
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "poll(int,int,int)",
                               "event: " + event);
                }
                if (event instanceof DoneEvent) {
                    if (((DoneEvent) event).id == id) {
                        done = true;
                    } else {
                        eventStack.add(event);
                    }
                } else if (event instanceof InputEvent) {
                    if ((discard & INPUT_EVENTS) != 0) {
                        ((InputEvent) event).consume();
                    } else if ((delay & INPUT_EVENTS) != 0) {
                        eventStack.add(event);
                    } else {
                        try {
                            doDispatch(orig, event);
                        } catch (RuntimeException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("poll(int,int,int)", ex);
                            }
                        }
                    }
                } else if (event instanceof PaintEvent) {
                    if ((discard & PAINT_EVENTS) != 0) {
                        // ignore
                    } else if ((delay & PAINT_EVENTS) != 0) {
                        eventStack.add(event);
                    } else {
                        try {
                            doDispatch(orig, event);
                        } catch (RuntimeException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("poll(int,int,int)", ex);
                            }
                        }
                    }
                } else {
                    if ((discard & OTHER_EVENTS) == OTHER_EVENTS) {
                        // ignore
                    } else if ((delay & OTHER_EVENTS) == OTHER_EVENTS) {
                        eventStack.add(event);
                    } else {
                        try {
                            doDispatch(orig, event);
                        } catch (RuntimeException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("poll(int,int,int)", ex);
                            }
                        }
                    }
                }
            } catch (InterruptedException ex) {
            }
        }
        // put the other DoneEvents back on the Queue
        if (eventStack.size() >0) {
            for (Iterator i = eventStack.iterator(); i.hasNext();) {
                AWTEvent pevent = (AWTEvent) i.next();
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "poll(int,int,int)",
                               "pevent: " + pevent);
                }
                queue.postEvent(pevent);
            }
        }
    }

    // }}}
    // {{{ doDispatch

    /**
     * Helper method to dispatch one event.
     *
     * @param   orig    The original event.
     * @param   event   The same as orig or the true event, if orig is a
     *                  WrapperEvent.
     */
    private static void doDispatch(AWTEvent orig, AWTEvent event)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "doDispatch(AWTEvent,AWTEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "orig: " + orig + ", " +
                       "event: " + event);
        }
        if (orig instanceof EventQueue.WrapperEvent) {
            ((EventQueue.WrapperEvent) orig).dispatch();
        } else if (EventQueue.isInstalled()) {
            ((EventQueue) EventQueue.instance()).dispatchEvent(event);
        } else {
            Object src = event.getSource();
            if (event instanceof ActiveEvent) {
                ((ActiveEvent)event).dispatch();
            } else if (src instanceof Component) {
                ((Component)src).dispatchEvent(event);
            } else if (src instanceof MenuComponent) {
                ((MenuComponent)src).dispatchEvent(event);
            }
        }
    }


    // }}}
    // {{{ class DoneEvent

    /**
     * Helper Event that signals the end of AWT Event polling.
     */
    private static class DoneEvent extends AWTEvent
    {
        /**
         * Dummy source for the event.
         */
        private static Label dummy = new Label();

        /**
         * Id of the event.
         */
        public int id;

        /**
         * Create a new DoneEvent.
         *
         * @param       id      The id of the event.
         */
        public DoneEvent(int id)
        {
            super(dummy, AWTEvent.RESERVED_ID_MAX);
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "DoneEvent(int)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "id: " + id);
            }
            this.id = id;
        }

        /**
         * Get a String representation for the DoneEvent.
         */
        public String toString()
        {
            return "DoneEvent[id=" + id + "]";
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Keymaps
    //----------------------------------------------------------------------
    // {{{ prepareKeymaps

    /**
     * Prepare Keymaps of JTextField and JPasswordField, so that the
     * return key invokes a possible default button. This can be undone for
     * individual TextComponents with {@link #setDefaultEnterBinding
     * setDefaultEnterBinding}.
     */
    public static void prepareKeymaps()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "prepareKeymaps()", "");
        }
        if (defaultKeymap == null) {
            JTextField tf = new JTextField ();
            Keymap tfKeymap = tf.getKeymap();
            KeyStroke enter = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
            Action defaultEnterBinding = tfKeymap.getAction(enter);
            tfKeymap.removeKeyStrokeBinding(enter);
            new JPasswordField ().getKeymap().removeKeyStrokeBinding(enter);
            defaultKeymap = tf.addKeymap("old", tfKeymap);
            defaultKeymap.addActionForKeyStroke(enter, defaultEnterBinding);
        }
    }

    // }}}
    // {{{ setDefaultEnterBinding

    /**
     * Restore the default binding for the return key, if it was changed
     * by {@link #prepareKeymaps prepareKeymaps}.
     *
     * @param   text    The TextComponent for which to restore the binding.
     */
    public static void setDefaultEnterBinding(JTextComponent text)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setDefaultEnterBinding(JTextComponent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "text: " + text);
        }
        text.setKeymap(defaultKeymap);
    }

    // }}}

    //----------------------------------------------------------------------
    // Close windows on escape
    //----------------------------------------------------------------------
    // {{{ closeOnEscape

    /**
     * Add a KeyboardListener to a Window that posts a WINDOW_CLOSING event
     * on it when Escape is pressed. Useful for modal dialogs.
     *
     * @param   win
     */
    public static void closeOnEscape(final Window win)
    {
        win.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == e.VK_ESCAPE && e.getModifiers() == 0) {
                    Toolkit.getDefaultToolkit().getSystemEventQueue()
                        .postEvent(new WindowEvent
                                   (win,
                                    WindowEvent.WINDOW_CLOSING));
                    e.consume();
                }
            }
        });
    }

    // }}}

    //----------------------------------------------------------------------
    // Tree Scroll bugfix
    //----------------------------------------------------------------------
    // {{{ constrainScroll

    /**
     * Constrain the scrolling range of a {@link javax.swing.JScrollPane
     * JScrollPane} so that it never scrolls to a negative view position. This
     * helps fix a bug in the Swing 1.1.1 action for PageDown in a {@link
     * javax.swing.JTree JTree} that places the tree at the bottom of the
     * JScrollPane if its size is less than the visible area. <p>
     *
     * This method replaces the JScrollPane's {@link javax.swing.JViewport
     * JViewport} with one that checks the argument to {@link
     * javax.swing.JViewport#setViewPosition setViewPosition}.
     *
     * @param   scrollPane      The JScrollPane to adapt.
     */
    public static void constrainScroll(JScrollPane scrollPane)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "constrainScroll(JScrollPane)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "scrollPane: " + scrollPane);
        }

        ConstrainedViewport cv = new ConstrainedViewport ();
        JViewport vp = scrollPane.getViewport();
        Point pos = vp.getViewPosition();
        cv.setView(vp.getView());
        cv.setViewPosition(pos);
        scrollPane.setViewport(cv);
    }

    // }}}
    // {{{ class ConstrainedViewport

    /**
     * This class extends JViewport to restrict the arguments passed to
     * {@link JViewport#setViewPosition setViewPosition} to nonnegative
     * coordinates.
     */
    public static class ConstrainedViewport extends JViewport
    {
        /**
         * Create a new ConstrainedViewport.
         */
        public ConstrainedViewport ()
        {
        }

        /**
         * Sets the view coordinates that appear in the upper left
         * hand corner of the viewport, does nothing if there's no view.
         *
         * @param p  a Point object giving the upper left coordinates
         */
        public void setViewPosition(Point p)
        {
            Point p2 = new Point ( p.x < 0 ? 0 : p.x, p.y < 0 ? 0: p.y);
            super.setViewPosition(p2);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Popup helper
    //----------------------------------------------------------------------
    // {{{ showPopup

    /**
     * Bring up a JPopupMenu but make sure that it is fully visible on the
     * screen (if at all possible). Optionally it can be constrained to its
     * parent Window as well.
     *
     * @param   menu    The JPopupMenu to show.
     * @param   parent  The parent Component of the menu.
     * @param   x       The targeted x coordinate.
     * @param   y       The targeted y coordinate.
     * @param   constrain Whether the menu should be constrained to its parent
     *                  window.
     */
    public static void showPopup(JPopupMenu menu, Component parent,
                                 int x, int y, boolean constrain)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "showPopup(JPopupMenu,Component,int,int,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "menu: " + menu + ", " +
                       "parent: " + parent + ", " +
                       "x: " + x + ", " +
                       "y: " + y + ", " +
                       "constrain: " + constrain);
        }

        if (parent == null) {
            // not our job
            menu.show(parent, x, y);
            return;
        }

        menu.pack();
        Dimension size = menu.getPreferredSize();
        Window win;
        Point pos;
        if (parent instanceof Window) {
            win =  (Window) parent;
            pos = new Point (x, y);
        } else {
            win = SwingUtilities.windowForComponent(parent);
            pos = SwingUtilities.convertPoint(parent, x, y, win);
            x = pos.x;
            y = pos.y;
        }
        Point wPos = win.getLocation();

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "x: " + x + ", y: " + y + ", wPos: " + wPos +
                       ", size: " + size);
        }
        if (constrain) {
            Dimension wSize = win.getSize();

            if (x + size.width > wSize.width) {
                x = wSize.width - size.width;
            }
            if (x < 0) {
                x = 0;
            }
            if (y + size.height > wSize.height) {
                y = wSize.height - size.height;
            }
            if (y < 0) {
                y = 0;
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG,
                           "showPopup(JPopupMenu,Component,int,int,boolean)",
                           "Window constrain x: " + x + ", y: " + y);
            }
        }
        Dimension sSize = Toolkit.getDefaultToolkit().getScreenSize();
        x += wPos.x;
        y += wPos.y;
        if (x + size.width > sSize.width) {
            x = sSize.width - size.width;
        }
        if (x < 0) {
            x = 0;
        }
        if (y + size.height > sSize.height) {
            y = sSize.height - size.height;
        }
        if (y < 0) {
            y = 0;
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "Screen constrain x: " + x + ", y: " + y);
        }
        pos = SwingUtilities.convertPoint(win, x - wPos.x, y -wPos.y, parent);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "Converted back x: " + pos.x + ", y: " + pos.y);
        }
        menu.show(parent, pos.x, pos.y);
    }

    // }}}

    //----------------------------------------------------------------------
    // Scrolling
    //----------------------------------------------------------------------
    // {{{ scrollUpLine

    /**
     * Scroll the contents of a scrollPane up by one line if possible. Note
     * that this action typically corresponds to pressing the DOWN arrow key.
     *
     * @param   scrollPane      The scroll pane to scroll.
     */
    public static void scrollUpLine(JScrollPane scrollPane)
    {
        JViewport vp = scrollPane.getViewport();
        Component view = vp == null ? null : vp.getView();
        if (! (view instanceof Scrollable)) {
            return;
        }

        Rectangle visRect = vp.getViewRect();
        Dimension vSize = view.getSize();
        int amount = ((Scrollable)view).getScrollableUnitIncrement
            (visRect, SwingConstants.VERTICAL, -1);
        visRect.y += amount;
        if (visRect.y + visRect.height > vSize.height) {
            visRect.y = Math.max(0, vSize.height - visRect.height);
        }
        vp.setViewPosition(visRect.getLocation());
    }

    // }}}
    // {{{ scrollDownLine

    /**
     * Scroll the contents of a scrollPane down by one line if possible. Note
     * that this action typically corresponds to pressing the UP arrow key.
     *
     * @param   scrollPane      The scroll pane to scroll.
     */
    public static void scrollDownLine(JScrollPane scrollPane)
    {
        JViewport vp = scrollPane.getViewport();
        Component view = vp == null ? null : vp.getView();
        if (! (view instanceof Scrollable)) {
            return;
        }

        Rectangle visRect = vp.getViewRect();
        int amount = ((Scrollable)view).getScrollableUnitIncrement
            (visRect, SwingConstants.VERTICAL, -1);
        visRect.y -= amount;
        if (visRect.y < 0) {
            visRect.y = 0;
        }
        vp.setViewPosition(visRect.getLocation());
    }

    // }}}

    //----------------------------------------------------------------------
    // Cleanup
    //----------------------------------------------------------------------
    // {{{ cleanup

    /**
     * Break up a Component hierarchy to increase chances for garbage
     * collection. Swing has quite a few static caches that can delay or
     * even prevent garbage collection. This method tries to remove as many
     * references as possible.
     *
     * @param   c       The root of the hierarchy to break up.
     */
    public static void cleanup(Component c)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "cleanup(Component)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "c: " + c);
        }
        if (c.getParent() != null) {
            c.getParent().remove(c);
        }
        Component [] children = null;
        if (c instanceof JMenu) {
            children = ((JMenu) c).getMenuComponents();
        } else
        if (c instanceof Container) {
            children = ((Container) c).getComponents();
        }
        if (children != null) {
            for (int i = 0; i < children.length; i++) {
                ((Container) c).remove(children[i]);
                cleanup(children[i]);
            }
        }
        if (c instanceof JMenu) {
            ((JMenu) c).removeAll();
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tree expand/collapse all
    //----------------------------------------------------------------------
    // {{{ expandAll

    /**
     * Expand a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to expand.
     * @param   path    Path to the node to start at.
     */
    public static void expandAll(JTree tree, TreePath path)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "expandAll(JTree,TreePath)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path);
        }

        Object node = path.getLastPathComponent();
        TreeModel model = tree.getModel();
        if (model.isLeaf(node)) {
            return;
        }

        tree.expandPath(path);

        int num = model.getChildCount(node);
        for (int i = 0; i < num; i++) {
            expandAll(tree, path.pathByAddingChild(model.getChild(node, i)));
        }
    }

    // }}}
    // {{{ collapseAll

    /**
     * Collapse a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to collapse.
     * @param   path    Path to the node to start at.
     */
    public static void collapseAll(JTree tree, TreePath path)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "expandAll(JTree,TreePath)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path);
        }

        Object node = path.getLastPathComponent();
        TreeModel model = tree.getModel();
        if (model.isLeaf(node)) {
            return;
        }

        //  if (tree.isExpanded(path)
            //  || tree.hasBeenExpanded(path)) {
            int num = model.getChildCount(node);
            for (int i = 0; i < num; i++) {
                collapseAll(tree,
                            path.pathByAddingChild(model.getChild(node, i)));
            }
            tree.collapsePath(path);
        //  }
    }

    // }}}

    //----------------------------------------------------------------------
    // Auto size table column
    //----------------------------------------------------------------------
    // {{{ autoSizeTableColumn

    /**
     * Resize a table column so its fits exactly the broadest of its visible
     * cells. This will set the columns width as well as its preferred width.
     * Only works if the table's parent is a JViewport, i.e. the table is
     * nested in a JScrollPane.
     *
     * @param   table   The table that contains the column.
     * @param   column  The colunm index in table coordinates.
     */
    public static void autoSizeTableColumn(JTable table, int column)
    {
        if (table.getRowCount() == 0
            || ! (table.getParent() instanceof JViewport)) {
            return;
        }

        JViewport vp = (JViewport) table.getParent();
        Rectangle rect = vp.getViewRect();

        int start = table.rowAtPoint(new Point(0, rect.y));
        int end = table.rowAtPoint(new Point(0, rect.y + rect.height));
        if (end < 0) {
            end = table.getRowCount() - 1;
        }

        int size = 0;
        for (int i = start; i <= end; i++) {
            Component com = table.prepareRenderer
                (table.getCellRenderer(i, column), i, column);
            if (com instanceof JComponent) {
                size = Math.max(size,
                                ((JComponent) com).getPreferredSize().width);
            } else {
                size = Math.max(size, com.getSize().width);
            }
        }

        int ccol = table.convertColumnIndexToModel(column);
        table.getColumnModel().getColumn(ccol).setPreferredWidth(size);
        table.getColumnModel().getColumn(ccol).setWidth(size);
        table.revalidate();
    }

    // }}}

    //----------------------------------------------------------------------
    // Miscellaneous
    //----------------------------------------------------------------------
    // {{{ centerWindow

    /**
     * Place a window in the center of the screen.
     *
     * @param   win     The window
     *
     * @since   0.96.0
     */
    public static void centerWindow(Window win)
    {
        Dimension sSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension wSize = win.getSize();

        win.setLocation((sSize.width - wSize.width) / 2,
                        (sSize.height - wSize.height) / 2);
    }

    // }}}
}


