// {{{ copyright

/********************************************************************
 *
 * $Id: TableModelAdapter.java,v 1.6 2000/06/06 17:00:03 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

import java.util.Vector;

import javax.swing.table.TableModel;
import javax.swing.event.TableModelListener;
import javax.swing.event.TableModelEvent;

/**
 * Built after the Adapter pattern, this class acts as the base for
 * manipulators of {@link javax.swing.table.TableModel TableModels} which
 * leave the original model undisturbed. <p>
 *
 * It implements the TableModel by wrapping itself around an existing
 * TableModel and delegating all calls to it. <p>
 *
 * Additionally the {@link javax.swing.event.TableModelListener
 * TableModelListener} interface in implemented in order to intercept {@link
 * javax.swing.event.TableModelEvent TableModelEvents}. This mechnism can be
 * extended by derived classes to change the behaviour of the wrapped model.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.6 $
 */

public class TableModelAdapter
implements TableModel, TableModelListener
{
    // {{{ variables

    /**
     * The original TableModel used for delegation.
     */
    protected TableModel model;

    /**
     * The TableModelListeners registered with this model.
     */
    protected Vector listeners = new Vector();

    // }}}

    // {{{ constructors

    /**
     * Create a new TableModelAdapter
     *
     * @param   model   The TableModel to wrap.
     */
    public TableModelAdapter (TableModel model)
    {
        this.model = model;
        // Register with the wrapped model to enable forwarding
        // of TableModelEvents to the registered TableModelListeners
        model.addTableModelListener(this);
    }

    // }}}

    // {{{ getModel

    /**
     * Get the wrapped model.
     *
     * @return  The wrapped TableModel
     */
    public TableModel getModel()
    {
        return model;
    }

    // }}}
    // {{{ setModel

    /**
     * Set the wrapped model.
     *
     * @param   model   The new TableModel.
     */
    public void setModel(TableModel model)
    {
        if (this.model != null) {
            this.model.removeTableModelListener(this);
        }
        this.model = model;
        model.addTableModelListener(this);
        // notify full structural change
        tableChanged(new TableModelEvent (this, TableModelEvent.HEADER_ROW));
    }

    // }}}

    // ----------------------------------------------------------------------
    // the TableModel interface
    // ----------------------------------------------------------------------
    // {{{ getValueAt

    /**
     * Return an attribute value for a cell.
     *
     * @param   row     The row whose value is to be looked up.
     * @param   column  The column whose value is to be looked up.
     *
     * @return  The value Object at the specified cell.
     */
    public Object getValueAt(int row, int column)
    {
        return model.getValueAt(row, column);
    }


    // }}}
    // {{{ setValueAt

    /**
     * Set the attribute value for a cell.
     *
     * @param   value   The value to set.
     * @param   row     The row whose value is to be set.
     * @param   column  The column whose value is to be set.
     */
    public void setValueAt(Object value, int arow, int column)
    {
        model.setValueAt(value, arow, column);
    }

    // }}}
    // {{{ getRowCount

    /**
     * Get the number of rows in the table.
     *
     * @return  The number of rows.
     */
    public int getRowCount()
    {
        return model.getRowCount();
    }

    // }}}
    // {{{ getColumnCount

    /**
     * Get the number of columns in the table.
     *
     * @return  The number of columns.
     */
    public int getColumnCount()
    {
        return model.getColumnCount();
    }


    // }}}
    // {{{ getColumnName

    /**
     * Return the name of a column. This is used
     * to initialize the table's column header name.
     *
     * @param   column  The index of the column.
     *
     * @return  The name of the column.
     */
    public String getColumnName(int column)
    {
        return model.getColumnName(column);
    }

    // }}}
    // {{{ getColumnClass

    /**
     * Return the lowest common denominator Class in a column.
     * This is used by the table to set up a default renderer and
     * editor for the column.
     *
     * @param   column  The index of the column.
     *
     * @return  The common ancestor class of the object values in the model.
     */
    public Class getColumnClass(int column)
    {
        return model.getColumnClass(column);
    }


    // }}}
    // {{{ isCellEditable

    /**
     * Return true if the a cell is editable.
     * Otherwise, setValueAt() on the cell will not change the value of
     * that cell.
     *
     * @param   row     The row whose value is to be looked up.
     * @param   column  The column whose value is to be looked up.
     *
     * @return  True if the cell is editable.
     *
     * @see     #setValueAt(java.lang.Object, int, int)
     */
    public boolean isCellEditable(int row, int column)
    {
         return model.isCellEditable(row, column);
    }

    // }}}
    // {{{ addTableModelListener

    /**
     * Add a listener to the model.
     *
     * @param   listener        The TableModelListener to add.
     */
    public synchronized void
    addTableModelListener (TableModelListener listener)
    {
        listeners.addElement(listener);
    }

    // }}}
    // {{{ removeTableModelListener

    /**
     * Remove a listener from the model.
     *
     * @param   listener        The TableModelListener to remove.
     */
    public synchronized void
    removeTableModelListener (TableModelListener listener)
    {
        listeners.removeElement(listener);
    }

    // }}}

    // ----------------------------------------------------------------------
    // the TableModelListener interface
    // ----------------------------------------------------------------------
    // {{{ tableChanged

    /**
     * Called, when the wrapped model has changed in some way. Forward
     * the event to all registered TableModelListeners.
     *
     * @param   event   The event that happened.
     */
    public void tableChanged(TableModelEvent event)
    {
        Vector tmp;
        synchronized (this) {
            tmp = (Vector)listeners.clone();
        }
        for (int i = 0; i < tmp.size(); i++) {
            ((TableModelListener) tmp.elementAt(i)).tableChanged(event);
        }
    }

    // }}}
}
