// {{{ copyright

/********************************************************************
 *
 * $Id: WindowsTreeCellRenderer.java,v 1.6 2000/06/06 17:00:03 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JTree;
import javax.swing.UIManager;
import javax.swing.tree.DefaultTreeCellRenderer;

// }}}

/**
 * TreeCellRenderer that draws a Windows like focus border. This border is a
 * lot more conspicuous than the default focus border in all look and feels.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.6 $
 */
public class WindowsTreeCellRenderer
    extends DefaultTreeCellRenderer
{
    // {{{ variables

    /**
     * The default icon for the stipple border.
     */
    private static Icon defaultIcon = new ImageIcon
        (WindowsTreeCellRenderer.class.getResource
         ("/de/qfs/lib/resources/icons/stipple.gif"));

    /**
     * The icon used for the border.
     */
    private Icon icon;

    /**
     * The icon's width.
     */
    int iconWidth;

    /**
     * The icon's height.
     */
    int iconHeight;

    /**
     * Whether the focus border needs to be drawn.
     */
    private boolean hasFocus;

    /**
     * Whether the focus border is drawn around the icon.
     */
    boolean drawsFocusBorderAroundIcon;

    // }}}

    // {{{ WindowsTreeCellRenderer

    /**
     * Create a new WindowsTreeCellRenderer.
     */
    public WindowsTreeCellRenderer ()
    {
        icon = defaultIcon;
        if (icon != null) {
            iconWidth = icon.getIconWidth();
            iconHeight = icon.getIconHeight();
        }
        Object value = UIManager.get("Tree.drawsFocusBorderAroundIcon");
        drawsFocusBorderAroundIcon =
            (value != null && ((Boolean)value). booleanValue());
    }

    // }}}

    // {{{ getTreeCellRendererComponent

    /**
     * Get the Component that renders a tree node.
     *
     * @param   tree            The tree to draw.
     * @param   value           The tree node's value.
     * @param   selected        Whether the node is selected.
     * @param   expanded        Whether the node is expanded.
     * @param   leaf            Whether the node is a leaf.
     * @param   row             The node's row index.
     * @param   hasFocus        Whether the node is selected.
     *
     * @return  A Component to render the node.
     */
    public Component getTreeCellRendererComponent
    (JTree tree, Object value, boolean selected, boolean expanded,
     boolean leaf, int row, boolean hasFocus)
    {
        this.hasFocus = hasFocus;
        DefaultTreeCellRenderer com = (DefaultTreeCellRenderer)
            super.getTreeCellRendererComponent(tree, value, selected, expanded,
                                               leaf, row,
                                               hasFocus && icon == null);
        return com;
    }

    // }}}
    // {{{ setBorderIcon

    /**
     * Set the icon to use for painting the focus border.
     *
     * @param   icon    The icon for the border.
     */
    public void setBorderIcon(Icon icon)
    {
        this.icon = icon;
        iconWidth = icon.getIconWidth();
        iconHeight = icon.getIconHeight();
        if (iconWidth <= 0 || iconHeight <= 0) {
            // useless icon
            icon = null;
        }
    }

    // }}}

    // {{{ paint

    /**
     * Paint the node.
     *
     * @param   g       The graphics to paint on.
     */
    public void paint(Graphics g)
    {
        super.paint(g);
        if (hasFocus && icon != null) {
            int offset = 0;
            int ix = 0;
            int iy = 0;
            int width = getWidth();
            int height = getHeight();
            Graphics cg = g.create();

            if (drawsFocusBorderAroundIcon) {
                offset = 0;
            } else if (getIcon() != null) {
                offset = getIcon().getIconWidth() +
                    Math.max(0, getIconTextGap() - 1);
            }

            // top edge
            cg.setClip(offset, 0, width - offset, 1);
            for (ix = offset; ix < width; ix += iconWidth) {
                icon.paintIcon(this, cg, ix, 0);
            }

            // bottom edge
            cg.setClip(offset, height - 1, width - offset, 1);
            iy = (height - 1) - ((height - 1) % iconHeight);
            for (ix = offset; ix < width; ix += iconWidth) {
                icon.paintIcon(this, cg, ix, iy);
            }

            // left edge
            cg.setClip(offset, 1, 1, height - 2);
            for (iy = 0; iy < height; iy += iconHeight) {
                icon.paintIcon(this, cg, offset, iy);
            }

            // right edge
            cg.setClip(width - 1, 1, 1, height - 2);
            ix = width - 1 - ((width - 1 - offset) % iconWidth);
            for (iy = 0; iy < height; iy += iconHeight) {
                icon.paintIcon(this, cg, ix, iy);
            }
            cg.dispose();
        }
    }

    // }}}
}
