// {{{ copyright

/********************************************************************
 *
 * $Id: Logger.java,v 1.16 2000/07/04 13:39:54 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.log;

// {{{ imports

//  import java.rmi.RemoteException;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

// }}}

/**
 * This class simplyfies logging of messages. <p>
 *
 * It encapsulates the name of its owner and uses it as the sender of the
 * messages it logs. Additionally it provides an advanced scheme to reduce
 * unwanted logging, since logging can get rather expensive in terms of memory
 * and runtime consumption if used seriously. <p>
 *
 * Typical use of a Logger looks like this:
 * <pre>
 *  // Create the Logger for SomeClass
 *  private static Logger logger = new Logger (SomeCLass.class);
 *
 *  public SomeType someMethod (SomeOtherType someParam)
 *  {
 *      if (logger.level >= Log.MTD) {
 *          logger.log (Log.MTD, "someMethod(SomeOtherType)",
 *                      logger.level < Log.MTDDETAIL ? "" :
 *                      "someParam: " + someParam);
 *      }
 *      ...
 * }
 * </pre>
 *
 * That way, the log message for the method entry will only be constructed, if
 * logger.level is at least Log.MTD. If it is exactly Log.MTD, no parameters
 * will be added to the log message, if it is higher, the value of someParam
 * will be logged as well. <p>
 *
 * The value of logger.level can be customized through the {@link #setLogLevel
 * setLogLevel} et. al. methods. <p>
 *
 * Changes in the log levels can be monitored by adding a {@link
 * LogLevelListener LogLevelListener} via {@link #addLogLevelListener
 * addLogLevelListener}.
 *
 * @see Log
 * @author      Gregor Schmid
 * @version     $Revision: 1.16 $
 */

public class Logger
{
    // {{{ variables

    /**
     * Log level set indirectly by using {@link #setDefaultLogLevel
     * setDefaultLogLevel} etc. This attribute should be considered read-only.
     * It is public only to keep the clutter of protecting log calls to a
     * minimum.
     */
    public int level;

    /**
     * Default level for a Logger: Log.WRNDETAIL.
     */
    public final static int DEFAULT_LOG_LEVEL = Log.WRNDETAIL;

    /**
     * The name of the Class owning the Logger.
     */
    private String owner;

    /**
     * The log levels for the Loggers.
     */
    private static LogLevels levels = new LogLevels ();
    //  private static LogLevelsInterface levels;

    //  static {
        //  try {
            //  levels = (LogLevelsInterface)
                //  Class.forName("de.qfs.lib.log.LogLevels").newInstance();
        //  } catch (Exception ex) {
            //  // Can't create LogLevels, probably no RMI available.
            //  levels = new SimpleLogLevels ();
        //  }
    //  };

    // }}}

    // {{{ constructors

    /**
     * Construct a Logger owned by owner.
     *
     * @param   owner   The name of the owner.
     */
    public Logger (String owner)
    {
        this.owner = owner;
        level = getLogLevel(this);
    }

    /**
     * Construct a Logger owned by owner.
     *
     * @param   owner   The owner.
     */
    public Logger (Object owner)
    {
        this(owner.getClass().getName());
    }

    /**
     * Construct a Logger owned by owner.
     *
     * @param   owner   The class of the owner.
     */
    public Logger (Class owner)
    {
        this(owner.getName());
    }

    // }}}

    //----------------------------------------------------------------------
    // Log levels for classes
    //----------------------------------------------------------------------
    // {{{ setDefaultLogLevel

    /**
     * Set the default level for all Loggers, for which no level is set
     * either directly on the class or on (parts of) its package. This not
     * only affects future calls to {@link #getLogLevel getLogLevel}, but also
     * changes the log level of already registered loggers.
     *
     * @param   level   The new default level.
     */
    public static void setDefaultLogLevel(int level)
    {
        if (levels != null) {
            levels.setDefaultLogLevel(level);
        }
    }

    // }}}
    // {{{ setLogLevel

    /**
     * Set the log level for a class or package. If the name represents a
     * package or part of a package it must end with '.' to distinguish it
     * from a class name. Otherwise setting the level for e.g.
     * "de.qfs.lib.log.Log" would implicitly set the level for
     * "de.qfs.lib.log.Logger" as well.
     *
     * @param   name    The name of the class or package (part). Package names
     *                  must end with a dot ('.').
     * @param   level   The maximum level at which objects of the class or
     *                  classes from the package should log messages.
     */
    public static void setLogLevel(String name, int level)
    {
        if (levels != null) {
            levels.setLogLevel(name, level);
        }
    }

    // }}}
    // {{{ removeLogLevel

    /**
     * Remove the level that was set for a Class or package so the
     * Class or the package's Classes will now inherit the default log level
     * or a package (part) level.
     *
     * @param   name    The name of the class or package (part). Package names
     *                  must end with a dot.
     */
    public static void removeLogLevel(String name)
    {
        if (levels != null) {
            levels.removeLogLevel(name);
        }
    }

    // }}}
    // {{{ getLogLevel

    /**
     * Get the level for a Logger. Calling this method will register the
     * Logger requesting the level, so that later changes to the default
     * level, its package's level or explicitly for the class can change the
     * Logger's log level.
     *
     * @param   logger  The Logger requesting its log level.
     *
     * @return The log level for the class owning the logger, if it has been
     *         explicitly set with {@link #setLogLevel setLogLevel}, the level
     *         of its package or the default level set with {@link
     *         #setDefaultLogLevel setDefaultLogLevel} otherwise.
     */
    public static int getLogLevel(Logger logger)
    {
        if (levels != null) {
            return levels.getLogLevel(logger);
        }
        return DEFAULT_LOG_LEVEL;
    }

    // }}}
    // {{{ setLogLevels(Hashtable)

    /**
     * Set various levels by parsing a set of properties in a Hashtable. The
     * Hashtable's keys must be of the form "log-NAME" where "NAME" is the
     * name of a Class or package as required by {@link #setLogLevel
     * setLogLevel}. The corresponding value must be a String that parses as
     * an int that will be used as the level to set. <p>
     *
     * Key/value pairs that do not follow these conventions will be silently
     * ignored. Thus for example the system properties may be used to set log
     * levels as follows:
     * <pre>
     *     Logger.setLogLevels(System.getProperties());
     * </pre>
     *
     * The default level can be set with the key "log-default".<p>
     *
     * @param   properties      The Hashtable to parse.
     */
    public static void setLogLevels(Hashtable properties)
    {
        for (Enumeration e = properties.keys(); e.hasMoreElements(); ) {
            String name = (String) e.nextElement();
            if (name.startsWith("log-")) {
                try {
                    int level = Integer.parseInt
                        ((String) properties.get(name));
                    if (name.equals("log-default")) {
                        setDefaultLogLevel(level);
                    } else {
                        setLogLevel(name.substring(4), level);
                    }
                } catch (NumberFormatException ex) {
                } catch (ClassCastException ex) {
                }
            }
        }
    }

    // }}}
    // {{{ setLogLevels(ResourceBundle)

    /**
     * Set various levels by parsing a ResourceBundle. The ResourceBundle's
     * names must be of the form "log-NAME" where "NAME" is the name of a
     * Class or package as required by {@link #setLogLevel setLogLevel}. The
     * corresponding value must parse as an int and will be used as the level
     * to set. <p>
     *
     * Resources that do not follow these conventions will be silently
     * ignored.
     *
     * The default level can be set with the resource "log-default".<p>
     *
     * @param   rb      The ResourceBundle to parse.
     */
    public static void setLogLevels(ResourceBundle rb)
    {
        for (Enumeration e = rb.getKeys(); e.hasMoreElements(); ) {
            String name = (String) e.nextElement();
            if (name.startsWith("log-")) {
                try {
                    int level = Integer.parseInt(rb.getString(name));
                    if (name.equals("log-default")) {
                        setDefaultLogLevel(level);
                    } else {
                        setLogLevel(name.substring(4), level);
                    }
                } catch (MissingResourceException ex) {
                } catch (NumberFormatException ex) {
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // LogLevelListener related methods
    //----------------------------------------------------------------------
    // {{{ addLogLevelListener

    /**
     * Add a LogLevelListener for the log levels.
     *
     * @param   listener        The listener to add.
     */
    public static void addLogLevelListener(LogLevelListener listener)
    {
        levels.addLogLevelListener(listener);
    }

    // }}}
    // {{{ removeLogLevelListener

    /**
     * Remove a LogLevelListener for the log levels.
     *
     * @param   listener        The listener to remove.
     */
    public static void removeLogLevelListener(LogLevelListener listener)
    {
        levels.removeLogLevelListener(listener);
    }

    // }}}
    // {{{ getLogLevels

    /**
     * Get the current log levels. This method is needed to synchronize the
     * LogLevelListener after initialization.
     *
     * @return  An object array that contains an alternating sequence of
     *          class/package names and log levels. The levels are Integers
     *          that may be null for classes for which a level has been
     *          requested but not explicitly set.
     */
    public static Object[] getLogLevels()
    {
        return levels.getLogLevels();
    }

    // }}}
    // {{{ setLogLevel

    /**
     * Callback method for a LogLevelListener to change the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     * @param   level   The new log level.
     */
    public static void setLogLevel(LogLevelListener source,
                                   String name, int level)
    {
        levels.setLogLevel(source, name, level);
    }

    // }}}
    // {{{ removeLogLevel

    /**
     * Callback method for a LogLevelListener to remove the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     */
    public static void removeLogLevel(LogLevelListener source, String name)
    {
        levels.removeLogLevel(source, name);
    }

    // }}}

    //----------------------------------------------------------------------
    // log messages and exceptions
    //----------------------------------------------------------------------
    // {{{ log(int, String, String)

    /**
     * Log a message.
     *
     * @param   level   The level of the message.
     * @param   method  The method that sent the message.
     * @param   message The message.
     */
    public final void log(int level, String method, String message)
    {
        Log.log(level, System.currentTimeMillis(), owner,
                method, message);
    }

    // }}}
    // {{{ log(String, Throwable)

    /**
     * Log an Exception.
     *
     * @param   method  The method that sent the message.
     * @param   throwable The Throwable.
     */
    public final void log(String method, Throwable throwable)
    {
        Log.log(Log.ERR, owner, method, throwable);
    }

    // }}}
    // {{{ log(String, Throwable, String)

    /**
     * Log an Exception.
     *
     * @param   method  The method that sent the message.
     * @param   throwable The Throwable.
     * @param   detail  Some extra info to log with the Exception name.
     */
    public final void log(String method, Throwable throwable, String detail)
    {
        Log.log(Log.ERR, owner, method, throwable, detail);
    }

    // }}}
    // {{{ log(int, String, Throwable)

    /**
     * Log an Exception at some specified level.
     *
     * @param   level   The level for the Exception.
     * @param   method  The method that sent the message.
     * @param   throwable The Throwable.
     */
    public final void log(int level, String method, Throwable throwable)
    {
        Log.log(level, owner, method, throwable);
    }

    // }}}
    // {{{ log(int, String, Throwable, String)

    /**
     * Log an Exception at some specified level.
     *
     * @param   level   The level for the Exception.
     * @param   method  The method that sent the message.
     * @param   throwable The Throwable.
     * @param   detail  Some extra info to log with the Exception name.
     */
    public final void log(int level, String method, Throwable throwable,
                          String detail)
    {
        Log.log(level, owner, method, throwable, detail);
    }

    // }}}
    // {{{ dumpStack

    /**
     * Log a stack trace.
     *
     * @param   level   The level to log the stack trace at.
     * @param   method  The method that sent the message.
     * @param   msg     The message to log with the stack trace.
     */
    public void dumpStack(int level, String method, String msg)
    {
        try {
            throw new Exception (msg);
        } catch (Exception ex) {
            log(level, method, ex);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Others
    //----------------------------------------------------------------------
    // {{{ getOwnerName

    /**
     * Get the name of the Class owning this Logger.
     *
     * @return  The owner's name.
     */
    public String getOwnerName()
    {
        return owner;
    }

    // }}}
}
