// {{{ copyright

/********************************************************************
 *
 * $Id: RemoteLogLevelAdapter.java,v 1.2 2000/10/26 15:04:07 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.logrmi;

// {{{ imports

import java.rmi.RemoteException;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.log.LogLevelCallback;
import de.qfs.lib.log.LogLevelEvent;
import de.qfs.lib.log.LogLevelListener;

// }}}

/**
 * Adapter that connects a RemoteLogLevelListener to an application.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 */
public class RemoteLogLevelAdapter
    extends RemoteLogLevelCallbackImplBase
    implements LogLevelListener
{
    // {{{ variables

    /**
     * Maximum number of failures.
     */
    private final static int MAX_FAILURES = 3;

    /**
     * The listener to forward LogLevelEvents to.
     */
    RemoteLogLevelListener listener;

    /**
     * Number of failures when calling the remote listener. When this reaches
     * MAX_FAILURES, the listener is disconnected.
     */
    private int failures;

    // }}}

    // {{{ constructors

    /**
     * Create a new RemoteLogLevelAdapter.
     *
     * @throws  RemoteException
     */
    public RemoteLogLevelAdapter()
        throws RemoteException
    {
    }

    /**
     * Create a new RemoteLogLevelAdapter.
     *
     * @param   listener        The listener to forward LogLevelEvents to.
     *
     * @throws  RemoteException
     */
    public RemoteLogLevelAdapter(RemoteLogLevelListener listener)
        throws RemoteException
    {
        this.listener = listener;
    }

    // }}}

    // {{{ setRemoteListener

    /**
     * Set the listener to forward LogLevelEvents to.
     *
     * @param   listener        The listener to forward LogLevelEvents to.
     */
    public void setRemoteListener(RemoteLogLevelListener listener)
    {
        this.listener = listener;
        try {
            listener.setLogLevelCallback(this);
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogLevelCallback interface
    //----------------------------------------------------------------------
    // {{{ getLogLevels

    /**
     * Get the current log levels from the callback. This method is needed to
     * synchronize the LogLevelListener and the LogLevelCallback after the
     * connection has been established.
     *
     * @return  An object array that contains an alternating sequence of
     *          class/package names and log levels. The levels are Integers
     *          that may be null for classes for which a level has been
     *          requested but not explicitly set.
     */
    public Object[] getLogLevels()
    {
        return Logger.getLogLevels();
    }

    // }}}
    // {{{ setLogLevel

    /**
     * Callback method for a LogLevelListener to change the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     * @param   level   The new log level.
     */
    public void setLogLevel(String name, int level)
    {
        Logger.setLogLevel(this, name, level);
    }

    // }}}
    // {{{ removeLogLevel

    /**
     * Callback method for a LogLevelListener to remove the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     */
    public void removeLogLevel(String name)
    {
        Logger.removeLogLevel(this, name);
    }

    // }}}
    // {{{ getOutputLevel

    /**
     * Get the current output log level from the callback's application.
     *
     * @return  The current output log level.
     */
    public int getOutputLevel()
    {
        return Log.getOutputLevel();
    }

    // }}}
    // {{{ setOutputLevel

    /**
     * Set the output log level for the callback's application.
     *
     * @param level     Output log level to set.
     */
    public void setOutputLevel(int level)
    {
        Log.setOutputLevel(level);
    }

    // }}}
    // {{{ getPreQueueLevel

    /**
     * Return the current pre-queue level from the callback's application.
     *
     * @return  The current pre-queue level.
     */
    public int getPreQueueLevel()
    {
        return Log.getPreQueueLevel();
    }

    // }}}
    // {{{ setPreQueueLevel

    /**
     * Set the pre-queue level for the callback's application.
     *
     * @param level     The pre-queue level to set.
     */
    public void setPreQueueLevel(int level)
    {
        Log.setPreQueueLevel(level);
    }

    // }}}
    // {{{ isQueueing

    /**
     * Check whether the callback's application is using a log queue.
     *
     * @return  Whether the callback's application is queueing log entries.
     *
     * @see     Log#isQueueing
     */
    public boolean isQueueing()
    {
        return Log.isQueueing();
    }

    // }}}
    // {{{ setQueueing

    /**
     * Set whether the callback's application should use a log queue.
     *
     * @param   queue   Whether to queue log entries.
     *
     * @see     Log#setQueueing
     */
    public void setQueueing(boolean queue)
    {
        Log.setQueueing(queue);
    }

    // }}}
    // {{{ getQueueSize

    /**
     * Get the size of log queue of the callback's application.
     *
     * @return  The size of the log queue.
     *
     * @see     Log#getQueueSize
     */
    public int getQueueSize()
    {
        return Log.getQueueSize();
    }

    // }}}
    // {{{ setQueueSize

    /**
     * Set the size of log queue of the callback's application.
     *
     * @param   size    The size of the log queue.
     *
     * @see     Log#setQueueSize
     */
    public void setQueueSize(int size)
    {
        Log.setQueueSize(size);
    }

    // }}}
    // {{{ isDropOnOverflow

    /**
     * Check whether the callback's application is dropping entries when
     * its log queue overflows.
     *
     * @return  Whether entries are dropped on overflow.
     *
     * @see     Log#isDropOnOverflow
     */
    public boolean isDropOnOverflow()
    {
        return Log.isDropOnOverflow();
    }

    // }}}
    // {{{ setDropOnOverflow

    /**
     * Set whether the callback's application should drop entries when
     * its log queue overflows.
     *
     * @param   drop    Whether to drop entries on overflow.
     *
     * @see     Log#setDropOnOverflow
     */
    public void setDropOnOverflow(boolean drop)
    {
        Log.setDropOnOverflow(drop);
    }

    // }}}
    // {{{ getFlushBufferSize

    /**
     * Get the size of the flush buffer of the callback's application.
     *
     * @return  The flush buffer size or 0 to indicate no flush buffer.
     */
    public int getFlushBufferSize()
    {
        return Log.getFlushBufferSize();
    }

    // }}}
    // {{{ setFlushBufferSize

    /**
     * Set the size of the flush buffer of the callback's application. Setting
     * it to 0 will turn the flush buffer off.
     *
     * @param   size    The size of the flush buffer.
     */
    public void setFlushBufferSize(int size)
    {
        Log.setFlushBufferSize(size);
    }

    // }}}
    // {{{ getFlushTriggerLevel

    /**
     * Get the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @return  The flush buffer's trigger level.
     */
    public int getFlushTriggerLevel()
    {
        return Log.getFlushTriggerLevel();
    }

    // }}}
    // {{{ setFlushTriggerLevel

    /**
     * Set the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @param   level   The trigger level to set.
     */
    public void setFlushTriggerLevel(int level)
    {
        Log.setFlushTriggerLevel(level);
    }

    // }}}
    // {{{ getPostFlushSize

    /**
     * Get the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @return  The number of messages to pass.
     */
    public int getPostFlushSize()
    {
        return Log.getPostFlushSize();
    }

    // }}}
    // {{{ setPostFlushSize

    /**
     * Set the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @param   size    The number of messages to pass.
     */
    public void setPostFlushSize(int size)
    {
        Log.setPostFlushSize(size);
    }

    // }}}
    // {{{ disconnect

    /**
     * End the connection with the RemoteLogLevelListener.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void disconnect()
         throws RemoteException
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogLevelListener interface
    //----------------------------------------------------------------------
    // {{{ classAdded

    /**
     * Notify the listener that a Logger for a formerly unknown class has
     * requested its log level, adding the class to the log level tree.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void classAdded(LogLevelEvent event)
    {
        if (listener != null) {
            try {
                listener.classAdded(event);
            } catch (RemoteException ex) {
            }
        }
    }

    // }}}
    // {{{ levelChanged

    /**
     * Notify the listener that a log level has been changed.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void levelChanged(LogLevelEvent event)
    {
        if (listener != null) {
            try {
                listener.levelChanged(event);
            } catch (RemoteException ex) {
            }
        }

    }

    // }}}
    // {{{ levelRemoved

    /**
     * Notify the listener that a log level has been unset.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void levelRemoved(LogLevelEvent event)
    {
        if (listener != null) {
            try {
                listener.levelRemoved(event);
            } catch (RemoteException ex) {
            }
        }
    }

    // }}}
    // {{{ setLogLevelCallback

    /**
     * Sets the LogLevelCallback for the LogLevelListener.
     *
     * @param callback  The LogLevelCallback to use to change the log
     *                  levels.
     */
    public void setLogLevelCallback(LogLevelCallback callback)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ failed

    /**
     * Take care of RMI failures.
     */
    private void failed()
    {
        if (++failures >= MAX_FAILURES) {
            // it doesn't make any sense to try to disconnect from the listener
            listener = null;
        }
    }

    // }}}
}
