// {{{ copyright

/********************************************************************
 *
 * $Id: EditorOptionEdit.java,v 1.5 2001/02/04 12:19:51 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;

import javax.swing.JScrollPane;
import javax.swing.JEditorPane;
import javax.swing.UIManager;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * An OptionEdit for an {@link EditorOption EditorOption}. It is represented
 * by a {@link javax.swing.JEditorPane JEditorPane} and uses the following
 * resources: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.type</code></TD>
 * <TD>The content type for the editor (optional, default text/plain).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.font</code></TD>
 * <TD>The font for the edit (optional, default is standard TextField font).
 * </TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.weight</code></TD>
 * <TD>The weight (a double) to use for the edit when placed in the {@link
 * de.qfs.lib.gui.Grid Grid} of an {@link OptionGroup OptionGroup} (optional,
 * default 0).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.scroll</code></TD>
 * <TD>If true the edit will get its own scroll pane (optional, default
 * true).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 * @since       0.98.0
 */

// }}}
public class EditorOptionEdit
    extends SimpleOptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (EditorOptionEdit.class);

    /**
     * The y-weight the edit should get in a Grid.
     * @serial
     */
    private double weight;

    /**
     * The scroll pane for the TextOptionEdit.
     */
    private JScrollPane scroll;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ EditorOptionEdit

    /**
     * Create a new EditorOptionEdit. Don't call this directly, use {@link
     * OptionEdit#makeEdit OptionEdit.makeEdit} instead.
     */
    public EditorOptionEdit()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the OptionEdit.
     */
    public void init()
    {
        super.init();

        String type = Option.getResources().getString
            ("option." + option.getName() + ".type", "text/plain");

        text = new JEditorPane (type, (String) option.getValue()) {
            public boolean isFocusTraversable()
            {
                return isEditable() && this.isEnabled();
            }
        };
        label.setLabelFor(text);
        text.setName("Editor-" + option.getName());

        // Override some resources to get a unified look for Text and String
        // edits.
        text.setBorder(UIManager.getBorder("TextField.border"));
        text.setFont(UIManager.getFont("TextField.font"));

        boolean found = false;
        String fontname =  Option.getResources().getString
            ("option." + option.getName() + ".font", null);
        if (fontname != null) {
            String[] fonts = text.getToolkit().getFontList();
            for (int i = 0; i < fonts.length; i++) {
                if (fonts[i].equals(fontname)) {
                    found = true;
                    break;
                }
            }
        }
        if (found) {
            Font current = text.getFont();
            text.setFont(new Font(fontname, current.getStyle(),
                                  current.getSize()));
        }

        boolean hasScroll = Option.getResources().getBoolean
            ("option." + option.getName() + ".scroll", true);
        if (hasScroll) {
            scroll = new JScrollPane (text);
            scroll.setBorder(text.getBorder());
            text.setBorder(null);
            scroll.setPreferredSize(new Dimension (10, 10));
            add(scroll, BorderLayout.CENTER);
        } else {
            scroll = null;
            add(text, BorderLayout.CENTER);
        }

        if (option.getValue() != null) {
            text.setText((String) option.getValue());
            text.setCaretPosition(0);
        }

        String sWeight = Option.getResources()
            .getString("option." + option.getName() + ".weight", null);
        if (sWeight != null) {
            try {
                weight = new Double(sWeight).doubleValue();
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "init()", ex);
                }
            }
        }
        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}
    // {{{ getWeight

    /**
     * Get the y-weight for the EditorOptionEdit.
     */
    public double getWeight()
    {
        return weight;
    }

    // }}}
}
