// {{{ copyright

/********************************************************************
 *
 * $Id: Option.java,v 1.9 2000/09/18 16:04:50 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;

import java.io.Serializable;

import java.util.MissingResourceException;
import java.util.Observable;
import java.util.ResourceBundle;
import java.util.Vector;

import de.qfs.lib.gui.Message;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

/**
 * An Option is a value with a name and some possible constraints. Its main
 * intention is to be used as a user editable option in a GUI application, but
 * it is also useful for configuration in general. <p>
 *
 * Options can be grouped together into an {@link OptionSet OptionSet} and
 * edited with the help of an {@link OptionGroup OptionGroup}. A component to
 * edit a single Option can be created with {@link OptionEdit#makeEdit
 * OptionEdit.makeEdit}. <p>
 *
 * Options are serializable as long as their values are. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.9 $
 */
public class Option extends Observable
    implements Serializable, Cloneable
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    final static long serialVersionUID = 3380333241949153766L;

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (Option.class);

    /**
     * The resources for the <code>de.qfs.lib.option</code> package.
     */
    private static MapResourceBundle rb = new MapResourceBundle ();

    /**
     * The OptionValidators for the Option.
     */
    private transient Vector validators;

    /**
     * The name of the Option.
     *
     * @serial
     */
    private String name;

    /**
     * The value of the Option.
     *
     * @serial
     */
    private Object value;

    /**
     * Whether the Option's value may be null.
     *
     * @serial
     */
    private boolean nullOK;

    // try to initialize resources
    static {
        try {
            rb.fetchProperties("/de/qfs/lib/resources/properties/qflib",
                               Option.class);
        } catch (MissingResourceException ex) {
            // so be it
        }
    }

    // }}}

    // {{{ constructors

    /**
     * Create a new Option.
     */
    public Option ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "Option()", "");
        }
    }

    /**
     * Create a new Option with a name and a value.
     *
     * @param   name    The name to use.
     * @param   value   The value to use.
     */
    public Option (String name, Object value)
    {
        this(name, value, false);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "Option(String,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
    }

    /**
     * Create a new Option with a name and a value and the nullOK flag.
     *
     * @param   name    The name to use.
     * @param   value   The value to use.
     * @param   nullOK  True to allow null values.
     */
    public Option (String name, Object value, boolean nullOK)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "Option(String,Object,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value + ", " +
                       "nullOK: " + nullOK);
        }
        this.name = name;
        this.value = value;
        this.nullOK = nullOK;
    }

    // }}}

    //----------------------------------------------------------------------
    // The resources
    //----------------------------------------------------------------------
    // {{{ getResources

    /**
     * Get the resources for the <code>de.qfs.lib.option</code> package. See
     * the package documentation for details about the resource names used.
     *
     * @return  The resources for the package.
     */
    public final static MapResourceBundle getResources()
    {
        return rb;
    }

    // }}}
    // {{{ setResources

    /**
     * Set the resources of the <code>de.qfs.lib.option</code> package. See
     * the package documentation for details about the resource names used.
     *
     * @param   resources       The resources to set.
     */
    public final static void setResources(ResourceBundle resources)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setResources(ResourceBundle)", "");
        }
        if (resources == null) {
            rb = new MapResourceBundle ();
        } else if (rb instanceof MapResourceBundle) {
            rb = (MapResourceBundle) resources;
        } else {
            rb = new MapResourceBundle (resources);
        }
    }

    // }}}
    // {{{ addResources

    /**
     * Add some resources for the <code>de.qfs.lib.option</code> package. See
     * the package documentation for details about the resource names used.
     *
     * @param   resources       The resources to add.
     */
    public final static void addResources(ResourceBundle resources)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addResources(ResourceBundle)", "");
        }
        rb.addResources(resources);
    }

    // }}}

    //----------------------------------------------------------------------
    // Attributes
    //----------------------------------------------------------------------
    // {{{ getName

    /**
     * Get the name of the Option.
     *
     * @return  The name of the Option.
     */
    public String getName()
    {
        return name;
    }

    // }}}
    // {{{ setName

    /**
     * Set the name of the Option.
     *
     * @param   name    The name to set.
     */
    public void setName(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setName(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        this.name = name;
    }

    // }}}
    // {{{ getValue

    /**
     * Get the value of the Option.
     *
     * @return  The value of the Option.
     */
    public Object getValue()
    {
        return value;
    }

    // }}}
    // {{{ setValue

    /**
     * Set the value of the Option.
     *
     * @param   value   The value to set.
     */
    public void setValue(Object value)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setValue(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "value: " + value);
        }
        this.value = value;
        setChanged();
    }

    // }}}
    // {{{ isNullOK

    /**
     * Query whether null is a legal value for the Option.
     *
     * @return  True if null is OK.
     */
    public boolean isNullOK()
    {
        return nullOK;
    }

    // }}}
    // {{{ setNullOK

    /**
     * Set whether null is a legal value for the Option.
     *
     * @param   nullOK  True to allow null values.
     */
    public void setNullOK(boolean nullOK)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setNullOK(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "nullOK: " + nullOK);
        }
        this.nullOK = nullOK;
    }

    // }}}

    //----------------------------------------------------------------------
    // Checking
    //----------------------------------------------------------------------
    // {{{ checkValue

    /**
     * Check, whether the Option accepts a value. If it does not, the reason
     * should be explained in a message dialog unless component is null.
     *
     * @param   component       The Component in which the Option is being
     *                          edited.
     * @param   value           The value to check.
     *
     * @return  True, if the value is acceptable.
     */
    public boolean checkValue(Component component, Object value)
    {
        if (!nullOK && value == null) {
            if (component != null) {
                Message.showMessage(component, "option.error.empty");
            }
            return false;
        }
        if (validators != null) {
            for (int i = 0; i < validators.size(); i++) {
                OptionValidator validator = (OptionValidator)
                    validators.elementAt(i);
                if (! validator.checkValue(component, this, value)) {
                    return false;
                }
            }
        }
        return true;
    }

    // }}}
    // {{{ addValidator

    /**
     * Add an OptionValidator to the Option.
     *
     * @param   validator       The OptionValidator to add.
     */
    public void addValidator(OptionValidator validator)
    {
        if (validators == null) {
            validators = new Vector ();
        }
        validators.addElement(validator);
    }

    // }}}
    // {{{ removeValidator

    /**
     * Remove an OptionValidator from the Option.
     *
     * @param   validator       The OptionValidator to remove.
     */
    public void removeValidator(OptionValidator validator)
    {
        if (validators != null) {
            validators.removeElement(validator);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Standard methods
    //----------------------------------------------------------------------
    // {{{ toString

    /**
     * Get a String representation for the Option.
     *
     * @return  The String representation.
     */
    public String toString()
    {
        return "Option[name=" + name + ",value=" + value + "]";
    }


    // }}}
    // {{{ hashCode

    /**
     * An Option's hashCode is generated from its name and value.
     *
     * @return  The hashCode.
     */
    public int hashCode()
    {
        return name.hashCode() + (value == null ? 0 : value.hashCode());
    }

    // }}}
    // {{{ equals

    /**
     * Two Options are considered equal if their names and values are equal.
     *
     * @param   object  The object to compare to.
     *
     * @return  True if objects are equal
     */
    public boolean equals(Object object)
    {
        if (object instanceof Option) {
            Option option = (Option) object;
            if (name.equals(option.getName())
                && ((value == null && option.getValue() == null)
                    || value.equals(option.getValue()))) {
                return true;
            }
        }
        return false;
    }

    // }}}
    // {{{ clone

    /**
     * Create a shallow copy of the Option. The option's value is not cloned.
     *
     * @return  The copy.
     */
    public Object clone()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "clone()", "");
        }
        try {
            return (Option) super.clone();
        } catch (CloneNotSupportedException ex) {
            //can't happen
            return null;
        }
    }

    // }}}
}
