// {{{ copyright

/********************************************************************
 *
 * $Id: OptionEdit.java,v 1.13 2001/02/04 12:19:52 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;

import javax.swing.JPanel;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.ClassWrapper;

// }}}

/**
 * An OptionEdit is a component used to edit an {@link Option Option's}
 * value. <p>
 *
 * An {@link OptionGroup OptionGroup} can be used to group OptionEdits into a
 * larger component, which in turn can be used inside an {@link OptionDialog
 * OptionDialog}. The layout and labels of the OptionEdits and OptionGroups
 * are determined by a set of resources set with {@link Option#setResources
 * Option.setResources}. See the package documentation for details about the
 * resource names used. <p>
 *
 * OptionEdit uses a {@link de.qfs.lib.util.ClassWrapper ClassWrapper} to
 * create specialiced OptionEdits via {@link #makeEdit makeEdit}.  The AWT
 * component name of an OptionEdit created with makeEdit is initialized to
 * <code>"OptionEdit-&lt;name&gt;"</code> where <code>&lt;name&gt;</code> is
 * the name of the option being edited. It is up to the derived classes to set
 * expressive names on their child components in the style
 * <code>"Label-&lt;name&gt;"</code>, <code>"Text-&lt;name&gt;"</code>
 * etc. <p>
 *
 * You will rarely need to use this class directly, except for deriving a
 * spcialized OptionEdit for a custom Option of your own design.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.13 $
 */
public abstract class OptionEdit
    extends JPanel
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (OptionEdit.class);

    /**
     * The ClassWrapper that creates the OptionEdits.
     */
    private static ClassWrapper wrapper =
        new ClassWrapper ("de.qfs.lib.option", "Edit", false);

    /**
     * Whether the edit is permanently disabled. If this flag is true,
     * {@link #setEnabled setEnabled} will be forced to disable the edit.
     */
    protected boolean disabled;

    /**
     * The Option to edit.
     */
    protected Option option;

    /**
     * The OptionGroup to which the edit belongs.
     */
    protected OptionGroup group;

    // }}}

    // {{{ constructor

    /**
     * Hidden constructor. Use {@link #makeEdit makeEdit} to create
     * OptionEdits.
     */
    protected OptionEdit ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // The factory
    //----------------------------------------------------------------------
    // {{{ makeEdit

    /**
     * A factory method that dynamically creates an OptionEdit for an Option.
     * The class of the edit component will be determined by adding "Edit" to
     * the (unqualified) class name of the Option. If you want to provide
     * OptionEdits for new Option types outside the
     * <code>de.qfs.lib.option</code> package you can do so by either having
     * your Option implement {@link de.qfs.lib.util.ClassWrapperAware
     * ClassWrapperAware} or by adding the package of your OptionEdit with
     * {@link #addPackage addPackage}. <p>
     *
     * After the OptionEdit instance is created its {@link #option option}
     * member is set and it is initialized through {@link #init init}.
     *
     * @param   option  The object to create an OptionEdit for.
     *
     * @return  The new OptionEdit or null if no matching class could be found.
     *
     * @see     de.qfs.lib.util.ClassWrapper
     */
    public static OptionEdit makeEdit(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeEdit(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        OptionEdit edit = null;

        if (option != null) {
            edit = (OptionEdit) wrapper.makeWrapper(option);
        }

        if (edit != null) {
            edit.option = option;
            edit.setName("OptionEdit-" + option.getName());
            edit.init();
        }

        return (OptionEdit) edit;
    }

    // }}}
    // {{{ addPackage

    /**
     * Add a package to be searched for OptionEdit classes, to enable
     * OptionEdits for user defined Options outside of the
     * <code>de.qfs.lib.option</code> package. Packages added later are
     * considered more specific and are searched first.
     *
     * @param   packagename     The name of the package to add.
     *
     * @see     de.qfs.lib.util.ClassWrapper#addPackage
     */
    public static void addPackage(String packagename)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addPackage(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "packagename: " + packagename);
        }
        wrapper.addPackage(packagename);
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getOption

    /**
     * Get the Option that is being edited by this OptionEdit.
     *
     * @return  The Option being edited.
     */
    public Option getOption()
    {
        return option;
    }

    // }}}
    // {{{ update(OptionSet)

    /**
     * Update the display of the edit component after a possible change of its
     * Option or the Option's value.
     *
     * @param   options The OptionSet containing the (possibly new) option.
     */
    public void update(OptionSet options)
    {
        if (option != null) {
            // Look for an Option with the same name
            Option opt = options.getOption(option.getName());
            if (opt != null) {
                update(opt);
            }
        }
    }

    // }}}
    // {{{ isDisabled

    /**
     * Query whether the edit is permanently disabled. In this case it will
     * force calls to {@link #setEnabled setEnabled} to disable the group.
     *
     * @return  True if the OptionGroup is disabled.
     * @since   0.98.0
     */
    public boolean isDisabled()
    {
        return disabled;
    }

    // }}}
    // {{{ setDisabled

    /**
     * Set whether theedit is permanently disabled. In this case it will force
     * calls to {@link #setEnabled setEnabled} to disable the group.
     *
     * @param   disabled        True to disable the group. If true,
     *                          {@link #setEnabled setEnabled(false)} is
     *                          called automatically.
     * @since   0.98.0
     */
    public void setDisabled(boolean disabled)
    {
        this.disabled = disabled;
        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}

    // {{{ getOptionGroup

    /**
     * Get the OptionGroup to which the edit belongs.
     *
     * @return  The edit's OptionGroup.
     * @since   0.98.0
     */
    public OptionGroup getOptionGroup()
    {
        return group;
    }

    // }}}
    // {{{ setOptionGroup

    /**
     * Set the OptionGroup to which the edit belongs.
     *
     * @param   group   The OptionGroup to set.
     * @since   0.98.0
     */
    public void setOptionGroup(OptionGroup group)
    {
        this.group = group;
    }


    // }}}

    // {{{ getWeight

    /**
     * Get the y-weight for the OptionEdit.
     *
     * @return  The default weight is 0.0, unless overridden in a derived
     *          class.
     *
     * @since   0.98.0
     */
    public double getWeight()
    {
        return 0.0;
    }

    // }}}

    //----------------------------------------------------------------------
    // Abstract methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the OptionEdit. This method is called by the {@link
     * #makeEdit makeEdit} factory method after the option member has been
     * set. <p>
     *
     * Must be implemented by a derived class.
     */
    protected abstract void init();

    // }}}
    // {{{ update(Option)

    /**
     * Update the display of the edit component after a possible change of its
     * Option or the Option's value. <p>
     *
     * Must be implemented by a derived class.
     *
     * @param   option The (possibly new) option.
     */
    public abstract void update(Option option);

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus. The OptionEdit should check the value
     * against its own constraints as well as pass it to {@link
     * Option#checkValue Option.checkValue}. <p>
     *
     * Must be implemented by a derived class.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public abstract boolean checkValue();

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option. <p>
     *
     * Must be implemented by a derived class.
     */
    public abstract void readValue();

    // }}}
}
