// {{{ copyright

/********************************************************************
 *
 * $Id: OptionGroup.jpp,v 1.20 2001/02/04 12:19:47 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Color;
import java.awt.Component;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


import java.util.StringTokenizer;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.UIManager;
import javax.swing.border.TitledBorder;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;
import de.qfs.lib.util.DynamicClassLoader;

// }}}

// {{{ doc

/**
 * OptionGroups form a hierarchie of nested GUI components in which Options
 * are grouped together for convenient editing. The hierarchie of the groups,
 * their layout and labels are determined by the resources set with {@link
 * Option#setResources Option.setResources}. See the package
 * documentation for details about the resource names used. <p>
 *
 * The preferred way to create an OptionGroup is through {@link #makeGroup
 * makeGroup} to create a new group, or {@link #getGroup getGroup} to reuse a
 * possibly existing group. <p>
 *
 * The AWT component displaying the OptionGroup can be accessed via {@link
 * #getComponent getComponent}. The component's name should be set by derived
 * classes to <code>"OptionGroup-&lt;name&gt;-&lt;options&gt;"</code> where
 * <code>&lt;name&gt;</code> is the resource name of the OptionGroup and
 * <code>&lt;options&gt;</code> is the configurable name of the OptionSet
 * being edited.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.20 $
 */

// }}}
public abstract class OptionGroup
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (OptionGroup.class);

    /**
     * The parent OptionGroup.
     */
    protected OptionGroup parent;

    /**
     * The members of the OptionGroup. May only contain OptionGroups and
     * {@link OptionEdit OptionEdits}.
     */
    protected List members;

    /**
     * The name of the OptionGroup.
     */
    protected String name;

    /**
     * The options to group.
     */
    protected OptionSet options;

    /**
     * The TitledBorder of the Frame.
     */
    protected TitledBorder titleBorder;

    /**
     * The title of the FramedOptionGroup's border.
     */
    protected String title;

    /**
     * Whether the group is permanently disabled. If this flag is true,
     * {@link #setEnabled setEnabled} will be forced to disable the group.
     */
    protected boolean disabled;

    /**
     * The current enabled state of the group.
     */
    protected boolean enabled = true;

    /**
     * Map group names to groups for reuse.
     */
    private static Map groups = new HashMap ();

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ OptionGroup()

    /**
     * No arg constructor for specialized OptionGroups.
     */
    protected OptionGroup ()
    {
    }

    // }}}
    // {{{ OptionGroup(String,OptionSet)

   /**
     * Create a new OptionGroup.
     *
     * @param   name    The name of the group.
     * @param   options The options to edit.
     */
    public OptionGroup(String name, OptionSet options)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "OptionGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "options: " + options);
        }
        this.name = name;
        this.options = options;
        members = new ArrayList();
        disabled = Option.getResources().getBoolean
            ("option.group." + name + ".disabled", false);
    }

    // }}}

    //----------------------------------------------------------------------
    // Factory methods
    //----------------------------------------------------------------------
    // {{{ getGroup

    /**
     * Get an OptionGroup for a set of Options. If a Group by that name
     * has already been created, it will be reused and updated, otherwise a
     * new OptionGroup is created.
     *
     * @param   name    The name of the group.
     * @param   options The options to edit.
     *
     * @return  The new OptionGroup.
     */
    public static OptionGroup getGroup(String name, OptionSet options)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getGroup(String,OptionGroup)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        OptionGroup group = (OptionGroup) groups.get(name);
        if (group == null) {
            group = makeGroup(name, options);
            if (group != null) {
                groups.put(name, group);
            }
        } else {
            group.update(options);
        }
        return group;
    }

    // }}}
    // {{{ makeGroup

    /**
     * Create a new OptionGroup. The type of the new group is determined by
     * the resources set with {@link Option#setResources
     * Option.setResources}. See the package documentation for details about
     * the resource names used. <p>
     *
     * A special case is type <code>custom</custom>. The class of the
     * OptionGroup to create is defined by the
     * <code>option.group.<em>groupname</em>.class</code> resource. The class
     * must have a public constructor of the form <br>
     *
     * <code>public CustomOptionGroup (String name, OptionSet options)</code>
     *
     * @param   name    The name of the group.
     * @param   options The options to edit.
     *
     * @return  The new OptionGroup.
     */
    public static OptionGroup makeGroup (String name, OptionSet options)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "options: " + options);
        }
        MapResourceBundle rb = Option.getResources();

        String type = rb.getString("option.group." + name + ".type", null);
        if (type == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "makeGroup(String,OptionSet)",
                           "No type for OptionGroup " + name);
            }
            return null;
        }
        if (type.equals("plain")) {
            return new PlainOptionGroup(name, options);
        }
        if (type.equals("frame")) {
            return new FramedOptionGroup(name, options);
        }
        if (type.equals("nested")) {
            return new NestedOptionGroup(name, options);
        }
        if (type.equals("radio")) {
            return new RadioOptionGroup(name, options);
        }
        if (type.equals("choice")) {
            return new ChoiceOptionGroup(name, options);
        }
        if (type.equals("tab")) {
            return new TabbedOptionGroup(name, options);
        }
        if (type.equals("custom")) {
            String clazz =
                rb.getString("option.group." + name + ".class", null);
            if (clazz == null) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "makeGroup(String,OptionSet)",
                               "No class for custom OptionGroup " + name);
                }
                return null;
            }
            try {
                Class custom;
                ClassLoader cl = OptionGroup.class.getClassLoader();
                if (cl == null) {
                    custom = Class.forName(clazz);
                } else {
                    custom = cl.loadClass(clazz);
                }
                Constructor cons = custom.getConstructor
                    (new Class[] {String.class, OptionSet.class});
                return (OptionGroup) cons.newInstance
                    (new Object[] {name, options});
            } catch (InvocationTargetException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "makeGroup(String,OptionSet)",
                               ex.getTargetException());
                    logger.log(Log.ERR, "makeGroup(String,OptionSet)", ex,
                               "name: " + name +
                               ", class: " + clazz);
                }
                return null;
            } catch (Exception ex) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "makeGroup(String,OptionSet)", ex,
                               "name: " + name +
                               ", class: " + clazz);
                }
                return null;
            }
        }
        if (logger.level >= Log.ERR) {
            logger.log(Log.ERR, "makeGroup(String,OptionSet)",
                       "Unknown OptionType " + type +
                       " for OptionGroup" + name);
        }
        return null;

    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getName

    /**
     * Get the name of the group.
     *
     * @return  The name of the group.
     * @since   0.98.0
     */
    public String getName()
    {
        return name;
    }

    // }}}
    // {{{ checkValues

    /**
     * Check whether all members of the group contain valid values.
     *
     * @return  True, if everything is OK.
     */
    public boolean checkValues()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "checkValues()", "");
        }
        for (Iterator i = members.iterator(); i.hasNext();) {
            Object member = i.next();
            if (member instanceof OptionGroup) {
                if (!((OptionGroup) member).checkValues()) {
                    return false;
                }
            } else {
                if (!((OptionEdit) member).checkValue()) {
                    return false;
                }
            }
        }
        return true;
    }

    // }}}
    // {{{ readValues

    /**
     * Read the values from the edit components and store them into the
     * Options. Should only be called after {@link #checkValues checkValues}
     * succeeded.
     */
    public void readValues()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValues()", "");
        }
        for (Iterator i = members.iterator(); i.hasNext();) {
            Object member = i.next();
            if (member instanceof OptionGroup) {
                ((OptionGroup) member).readValues();
            } else {
                ((OptionEdit) member).readValue();
            }
        }
    }

    // }}}
    // {{{ notifyOptionObservers

    /**
     * Notify the Observers of the Options being edited in this OptionGroup.
     */
    public void notifyOptionObservers()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "notifyOptionObservers()", "");
        }
        for (Iterator i = members.iterator(); i.hasNext();) {
            Object member = i.next();
            if (member instanceof OptionGroup) {
                ((OptionGroup) member).notifyOptionObservers();
            } else {
                ((OptionEdit) member).getOption().notifyObservers();
            }
        }
    }

    // }}}
    // {{{ update

    /**
     * Update display for possibly new Options. The name and type of
     * the Options in the OptionSet must be the same as before.
     *
     * @param   options The new OptionSet.
     */
    public void update(OptionSet options)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       " my name: " + name);
        }
        this.options = options;
        for (Iterator i = members.iterator(); i.hasNext();) {
            Object member = i.next();
            if (member instanceof OptionGroup) {
                ((OptionGroup) member).update(options);
            } else {
                ((OptionEdit) member).update(options);
            }
        }
    }

    // }}}
    // {{{ setEnabled

    /**
     * Set the enabled state for all members of the group.
     *
     * @param   enabled The enabled state to set.
     */
    public void setEnabled (boolean enabled)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setEnabled(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "enabled: " + enabled +
                       ", disabled: " + disabled);
        }

        if (disabled) {
            enabled = false;
        }

        if (members != null) {
            for (Iterator i = members.iterator(); i.hasNext();) {
                Object member = i.next();
                if (member instanceof OptionGroup) {
                    ((OptionGroup) member).setEnabled(enabled);
                }else {
                    ((OptionEdit) member).setEnabled(enabled);
                }
            }
        }

        if (titleBorder != null) {
            Color color;
            if (enabled) {
                color = UIManager.getColor("textText");
            } else {
                color = UIManager.getColor("textInactiveText");
            }
            titleBorder.setTitleColor(color);
            titleBorder.setBorder(BorderFactory.createLineBorder(color));
        }
    }

    // }}}
    // {{{ isDisabled

    /**
     * Query whether the OptionGroup is permanently disabled. In this case it
     * will force calls to {@link #setEnabled setEnabled} to disable the group.
     *
     * @return  True if the OptionGroup is disabled.
     */
    public boolean isDisabled()
    {
        return disabled;
    }

    // }}}
    // {{{ setDisabled

    /**
     * Set whether the OptionGroup is permanently disabled. In this case it
     * will force calls to {@link #setEnabled setEnabled} to disable the group.
     *
     * @param   disabled        True to disable the group. If true,
     *                          {@link #setEnabled setEnabled(false)} is
     *                          called automatically.
     */
    public void setDisabled(boolean disabled)
    {
        this.disabled = disabled;
        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}
    // {{{ setTitle

    /**
     * Set the title of the OptionGroup. Overrides the value read from the
     * .title property. Only useful if the group has a frame.
     *
     * @param   title   The title to set.
     * @since    0.98.0
     */
    public void setTitle(String title)
    {
        this.title = title;
        if (titleBorder != null) {
            titleBorder.setTitle(title);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Group hierarchy
    //----------------------------------------------------------------------
    // {{{ getParentGroup

    /**
     * Get the parent of the OptionGroup.
     *
     * @return  The OptionGroup's parent.
     * @since   0.98.0
     */
    public OptionGroup getParentGroup()
    {
        return parent;
    }

    // }}}
    // {{{ setParentGroup

    /**
     * Set the parent of the OptionGroup.
     *
     * @param   parent  The parent to set.
     * @since   0.98.0
     */
    public void setParentGroup(OptionGroup parent)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setParentGroup(OptionGroup)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "parent: " + parent);
        }
        this.parent = parent;
    }

    // }}}
    // {{{ getMemberCount

    /**
     * Get the number of members of the OptionGroup.
     *
     * @return  The number of the OptionGroup's members.
     * @since   0.98.0
     */
    public int getMemberCount()
    {
        return members.size();
    }

    // }}}
    // {{{ getMember

    /**
     * Get a member of the group.
     *
     * @param   index   The index of the member.
     *
     * @return  The member, either an OptionGroup or an OptionEdit.
     * @since   0.98.0
     */
    public Object getMember(int index)
    {
        return members.get(index);
    }

    // }}}
    // {{{ getIndexOfMember

    /**
     * Find a member of the OptionGroup and get its index.
     *
     * @param   member  The member to look for.
     *
     * @return  The member's index or -1.
     * @since   0.98.0
     */
    public int getIndexOfMember(Object member)
    {
        return members.indexOf(member);
    }

    // }}}
    // {{{ findGroup

    /**
     * Traverse the OptionGroup members and search for an OptionGroup.
     *
     * @param   name    The name of the group to search.
     *
     * @return  The group - if found - or null.
     *
     * @since   0.98.0
     */
    public OptionGroup findGroup(String name)
    {
        int cnt = getMemberCount();
        for (int i = 0; i < cnt; i++) {
            Object mem = getMember(i);
            if (mem instanceof OptionGroup) {
                if (name.equals(((OptionGroup) mem).getName())) {
                    return (OptionGroup) mem;
                }
                OptionGroup ret = ((OptionGroup) mem).findGroup(name);
                if (ret != null) {
                    return ret;
                }
            }
        }
        return null;
  }

    // }}}
    // {{{ findEdit

    /**
     * Traverse the OptionGroup members and search for an OptionEdit.
     *
     * @param   name    The option name of the edit to search.
     *
     * @return  The edit - if found - or null.
     *
     * @since   0.98.0
     */
    public OptionEdit findEdit(String name)
    {
        int cnt = getMemberCount();
        for (int i = 0; i < cnt; i++) {
            Object mem = getMember(i);
            if (mem instanceof OptionGroup) {
                OptionEdit ret = ((OptionGroup) mem).findEdit(name);
                if (ret != null) {
                    return ret;
                }
            } else if (mem instanceof OptionEdit) {
                if (name.equals(((OptionEdit) mem).getOption().getName())) {
                    return (OptionEdit) mem;
                }
            }
        }
        return null;
    }

    // }}}

    //----------------------------------------------------------------------
    // Abstract methods
    //----------------------------------------------------------------------
    // {{{ getComponent

    /**
     * Get the AWT Component for the OptionGroup.
     *
     * @return  The Component.
     */
    public abstract Component getComponent();

    // }}}

    //----------------------------------------------------------------------
    // Helpers for derived classes
    //----------------------------------------------------------------------
    // {{{ createMembers

    /**
     * Create the members of the OptionGroup.
     */
    protected void createMembers()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "createMembers()", "");
        }
        members = new ArrayList();
        MapResourceBundle rb = Option.getResources();

        String elems = rb.getString("option.group." + name + ".members", null);
        if (elems == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "createMembers()",
                           "No Members for OptionGroup " + name);
            }
            return;

        }
        if (elems.startsWith("sameas.")) {
            String ref = elems.substring("sameas.".length());
            elems = rb.getString("option.group." + ref + ".members", null);
            if (elems == null) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "createMembers()",
                               "No Members for referenced group " + ref);
                }
                return;
            }
        }

        StringTokenizer st = new StringTokenizer(elems, " ");
        for (; st.hasMoreTokens();) {
            String elem = st.nextToken();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "createMembers()",
                           "elem: " + elem);
            }
            if (elem.startsWith("group.")) {
                elem = elem.substring("group.".length());
                OptionGroup group = OptionGroup.makeGroup(elem, options);
                if (group != null) {
                    group.setParentGroup(this);
                    String enabled = rb.getString
                        ("option.group." + name + "." + elem + ".enabled",
                         "true");
                    if (! Boolean.valueOf(enabled).booleanValue()) {
                        group.setEnabled(false);
                    }
                    members.add(group);
                }
            } else {
                Option option = options.getOption(elem);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "createMembers()",
                               "option: " + option);
                }
                if (option == null) {
                    if (logger.level >= Log.ERR) {
                        logger.log(Log.ERR, "createMembers()",
                                   "unknown Option: " + elem);
                    }
                } else {
                    Component com = OptionEdit.makeEdit(option);
                    if (com == null) {
                        if (logger.level >= Log.ERR) {
                            logger.log(Log.ERR, "createMembers()",
                                       "Cannot create option edit for " +
                                       "Option " + option.getName());
                        }
                    } else {
                        ((OptionEdit) com).setOptionGroup(this);
                        String enabled = rb.getString
                            ("option.group." + name + "." +
                             option.getName() + ".enabled", "true");
                        if (! Boolean.valueOf(enabled).booleanValue()) {
                            com.setEnabled(false);
                        }
                        members.add(com);
                    }
                }
            }
        }
    }

    // }}}
    // {{{ makeBorder

    /**
     * Create the border for the group if the .frame property for the
     * group is set.
     *
     * @param   com     The component to place the border on.
     *
     * @since   0.98.0
     */
    protected void makeBorder(JComponent com)
    {
        if (! Option.getResources().getBoolean
            ("option.group." + name + ".frame", false)) {
            titleBorder = null;
            return;
        }
        Color color;
        if (enabled) {
            color = UIManager.getColor("textText");
        } else {
            color = UIManager.getColor("textInactiveText");
        }
        titleBorder = BorderFactory.createTitledBorder
            (BorderFactory.createLineBorder(color));
        titleBorder.setTitleColor(color);
        com.setBorder(BorderFactory.createCompoundBorder
                       (titleBorder,
                        BorderFactory.createEmptyBorder(0, 3, 3, 3)));
        if (title == null) {
            title = Option.getResources().getString
                ("option.group." + name + ".title", null);
        }
        if (title != null) {
            titleBorder.setTitle(title);
        }
    }

    // }}}
}
