// {{{ copyright

/********************************************************************
 *
 * $Id: PlainOptionGroup.java,v 1.12 2000/11/29 14:38:09 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;
import java.awt.Dimension;

import javax.swing.JComponent;

import de.qfs.lib.gui.Grid;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * A PlainOptionGroup packs Options and Subgroups together into a Grid. It
 * uses the following resources: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.type</code></TD>
 * <TD>A PlainOptionGroup is designated by the type <code>plain</code>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the group will be disabled and remain so even if
 * {@link OptionGroup#setEnabled setEnabled(true)} is called (optional,
 * default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.members</code></TD>
 * <TD>The members of the group, a space separated list where <em>name</em>
 * denotes an Option and group.<em>groupname</em> denotes an OptionGroup.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.columns</code></TD>
 * <TD>The number of columns in which to place the members of the group
 * (optional, default 1).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.frame</code></TD>
 * <TD>Whether the group should have a framed border (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.title</code></TD>
 * <TD>The title for a framed border of the group. Ignored if no frame is
 * specified(optional, default empty).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.12 $
 */

// }}}
public class PlainOptionGroup extends OptionGroup
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (PlainOptionGroup.class);

    /**
     * The Grid Component that holds the members of the PlainOptionGroup.
     */
    protected Grid grid;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ PlainOptionGroup()

    /**
     * No arg constructor for customized OptionGroups.
     */
    protected PlainOptionGroup ()
    {
    }

    // }}}
    // {{{ PlainOptionGroup(String,OptionSet)

    /**
     * Create a new PlainOptionGroup.
     *
     * @param   name    The name of the group.
     * @param   options The Options to edit.
     */
    public PlainOptionGroup (String name, OptionSet options)
    {
        super (name, options);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "PlainOptionGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        createMembers();
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getComponent

    /**
     * Get the AWT Component for the OptionGroup.
     *
     * @return  The Component.
     */
    public Component getComponent()
    {
        if (grid == null) {
            makeGrid();
        }
        return grid;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ makeGrid

    /**
     * Create the Grid component for the group members.
     */
    protected void makeGrid()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeGrid()", "");
        }
        String cols = Option.getResources()
            .getString("option.group." + name + ".columns", "1");
        int columns = Integer.parseInt(cols);

        grid = new Grid();
        grid.setName("OptionGroup-" + name + "-" +
                     options.getConfigurableName());
        for (int i = 0; i < members.size(); i+= columns) {
            int num = Math.min(columns, members.size() - i);
            Component[] coms = new Component [num];
            double weight = 0.0;
            for (int j = 0; j < num; j++) {
                Object mem = members.get(i + j);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "makeGrid()",
                               "member: " + mem);
                }

                if (mem instanceof OptionGroup) {
                    coms[j] = ((OptionGroup) mem).getComponent();
                    if (coms[j] instanceof Grid) {
                        weight = Math.max(weight,
                                          ((Grid) coms[j]).getTotalWeight());
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "makeGrid()",
                                       "weight: " + weight);
                        }
                    }
                } else {
                    coms[j] = (OptionEdit) mem;
                    weight = Math.max(weight, ((OptionEdit) mem).getWeight());
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "makeGrid()",
                                   "weight: " + weight);
                    }
                }
            }
            grid.addLine(coms, weight);
        }

        if (disabled) {
            setEnabled(false);
        }

        makeBorder(grid);
    }

    // }}}
}
