// {{{ copyright

/********************************************************************
 *
 * $Id: ArgsParser.jpp,v 1.10 2001/02/04 12:19:54 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;

import java.net.URL;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;


import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;


// }}}

// {{{ doc

/**
 * This class implements convenient command line parsing. It is not as
 * sophisticated as other command line parsers, e.g. the GNU getopt package,
 * but it is easy to use and has a straightforward interface. <p>
 *
 * An ArgsParser reads options and their parameters from an array of strings,
 * the standard format for command line arguments passed to an application's
 * main method. Additional support is provided for reading options from a
 * property file. <p>
 *
 * The options the ArgsParser expects can either be passed to the constructor,
 * or be set or modified with {@link #setOptions setOptions} or {@link
 * #addOptions addOptions}. <p>
 *
 * During parsing, all arguments starting with '-' are interpreted as
 * options. If an option has a required paramter, the following argument is
 * used, causing a MalformedArgumentListException if no further arguments are
 * given. If an option's parameter starts with '-', the option must be set
 * in the form '-option=-value'. <p>
 *
 * If an option has an optional paramter, the argument following an option is
 * interpreted as that option's paramter, if it does not start with '-',
 * otherwise the options paramter remains empty and the following argument is
 * considered to be another option. This kind of guessing is suppressed, if
 * the option's parameter indicator is '=' (see {@link #setOptions
 * setOptions}) <p>
 *
 * If an argument is of the form '-option=value' the option must have a
 * required or optional parameter and its name and value are extracted
 * accordingly. <p>
 *
 * Parsing stops when either the argument '--' is encountered, or an argument
 * that is not a parameter and does not start with '-' is encoutered. <p>
 *
 * Additional options can be read from a property file. This file can be
 * specified in three ways: directly via {@link #setPropertyFile
 * setPropertyFile}, through a system property, or via a command line
 * argument, in increasing order of precedence. The property to use can be set
 * with {@link #setProperty setProperty}, the name of the option to interpret
 * as source for extra options with {@link #setDefaultOption
 * setDefaultOption}. In all cases the source may either be a file name or a
 * valid URL. Arguments from the command line take precedence over options
 * read from the property file, unless the option is allowed to have multiple
 * values, in which case the values are combined. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.10 $
 */

// }}}
public class ArgsParser
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (ArgsParser.class);

    /**
     * The property file to read the options from.
     */
    private String propFile;

    /**
     * The system property specifying the file or URL for the default
     * options.
     */
    private String property;

    /**
     * The option that overrides the location for the default options.
     */
    private String propOption;

    /**
     * The options that are allowed.
     */
    private Hashtable options;

    /**
     * The wildcarded options.
     */
    private Vector wildcards;

    /**
     * The options that were parsed.
     */
    private Hashtable result;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ ArgsParser()

    /**
     * Create a new ArgsParser that will accept all options it encounters with
     * an optional parameter. Using this constructor is equivalent to calling
     * {@link #setOptions setOptions} with the single option definition '?*'.
     */
    public ArgsParser()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ArgsParser()", "");
        }
        options = new Hashtable ();
        wildcards = new Vector ();
        wildcards.addElement(new Option ("", '?', false, true));
    }

    // }}}
    // {{{ ArgsParser(String[])

    /**
     * Create a new ArgsParser, defining the options it should accept.
     *
     * @param   opts    The options to set. See {@link #setOptions setOptions}
     *                  for the format specification.
     *
     * @throws IllegalArgumentException If the option definitions are
     *                                  incorrect.
     */
    public ArgsParser(String[] opts)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ArgsParser(String[])", "");
        }
        setOptions(opts);
    }

    // }}}

    //----------------------------------------------------------------------
    // Defining the options
    //----------------------------------------------------------------------
    // {{{ setOptions

    /**
     * Set the options the ArgsParser should know about. Each String in the
     * array designates one option (or multiple similar ones) and must have
     * the following format:
     * <ul>
     * <li>Optional character 'm', indicating that the option may occur
     * multiple times on the command line. If it does, the resulting value
     * will be a string array containing all parameters instead of a single
     * string. If an option may not have multiple values but still occurs
     * more than once, the last occurancedefines the option's value.</li>
     * <li>Required parameter indicator, one of the characters '+', '-',
     * '?' or '=' with the following meaning: <br>
     * <strong>+</strong>: required parameter<br>
     * <strong>-</strong>: no parameter<br>
     * <strong>?</strong>: optional parameter<br>
     * <strong>=</strong>: optional parameter, no guessing<br>
     * The difference between '?' and '=' is that for '=' the ArgsParser only
     * reads a parameter if it is given in the <code>-name=value</code>
     * syntax. This is very useful for boolean options that normally don't
     * require a parameter. If a boolean option is set in a properties file
     * and should be negated on the commandline, either a negative option must
     * be defined, or the option must accept a parameter. However if the '?'
     * indicator is used, setting such an option as the last one just before
     * the rest of the arguments would require the '--' separator to avoid
     * interpreting the first non-option argument as the option's parameter
     * value.
     * </li>
     * <li>The name of the option</li>
     * <li>Optional wildcard '*', meaning all options starting with this name
     * are allowed. </li>
     *
     * @param   opts    The options to set.
     *
     * @throws IllegalArgumentException If the option definitions are
     *                                  incorrect.
     */
    public void setOptions(String[] opts)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setOptions(String[])", "");
        }
        options = new Hashtable ();
        wildcards = new Vector ();
        addOptions(opts);
    }

    // }}}
    // {{{ addOptions

    /**
     * Set some additional options the ArgsParser should know about.
     *
     * @param opts     The options to add. See {@link #setOptions setOptions}
     *                 for the format specification.
     *
     * @throws IllegalArgumentException If the option definitions are
     *                                  incorrect.
     */
    public void addOptions(String[] opts)
        throws IllegalArgumentException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addOptions(String[])", "");
        }

        for (int i = 0; i < opts.length; i++) {
            if (opts[i] == null || opts[i].length() < 2) {
                throw new IllegalArgumentException
                    ("Illegal option " + opts[i] + " at index " + i);
            }
            boolean multiple = opts[i].startsWith("m");
            boolean wildcard = opts[i].endsWith("*");
            char param = opts[i].charAt(multiple ? 1 : 0);
            String name = opts[i].substring
                (multiple ? 2 : 1, opts[i].length() - (wildcard ? 1 : 0));
            if (name.length() == 0 && !wildcard) {
                throw new IllegalArgumentException
                    ("Illegal option " + opts[i] + " at index " + i);
            }
            switch(param) {
            case '+':
            case '-':
            case '?':
            case '=':
                break;
            default:
                throw new IllegalArgumentException
                    ("Missing parameter identifier in option " + opts[i] +
                     " at index " + i);
            }
            Option option = new Option (name, param, multiple, wildcard);
            if (wildcard) {
                wildcards.addElement(option);
            } else {
                options.put(option.name, option);
            }
        }
    }

    // }}}
    // {{{ setDefault

    /**
     * Set a default value for an option. This must be called after
     * initializing all options with {@link #setOptions setOptions} or {@link
     * #addOptions addOptions} and before {@link #parse parse}. <p>
     *
     * For options with multiple values each default value will be part of the
     * result, independent of whether the option was set on the command line
     * or not.
     *
     * @param   name    The name of the option.
     * @param   value   The default value.
     *
     * @throws  UnknownOptionException  If an option hasn't been defined.
     * @throws  MissingParameterException       If a required parameter is
     *                  missing.
     * @throws  UnexpectedParameterException    If an unexpected parameter is
     *                  encountered.
     *
     * @since   0.98.0
     */
    public void setDefault(String name, String value)
        throws UnknownOptionException,
               MissingParameterException,
               UnexpectedParameterException
    {
        if (result == null) {
            result = new Hashtable ();
        }
        addResult(isAllowed(name), name, value);
    }

    // }}}

    //----------------------------------------------------------------------
    // Additional property files
    //----------------------------------------------------------------------
    // {{{ setPropertyFile

    /**
     * Set the file name or URL to read options from.
     *
     * @param   name    The file name or URL to set.
     */
    public void setPropertyFile(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setPropertyFile(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        propFile = name;
    }

    // }}}
    // {{{ setProperty

    /**
     * Set the system property name for the default options.
     *
     * @param   property        The property to set.
     */
    public final void setProperty(String property)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setProperty(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "property: " + property);
        }
        this.property = property;
    }

    // }}}
    // {{{ setDefaultOption

    /**
     * Set the option whose argument should be interpreted as source for
     * the default options. You should specify this option as requiring
     * an argument (see {@link #setOptions setOptions}).
     *
     * @param   option  The option name to set, without leading '-'.
     */
    public final void setDefaultOption(String option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setDefaultOption(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        propOption = option;
    }

    // }}}

    //----------------------------------------------------------------------
    // Parsing
    //----------------------------------------------------------------------
    // {{{ parse

    /**
     * Parse an array of command line options.
     *
     * @param   args    The arguments from the command line.
     *
     * @return  The index of the first non-option argument on the command line.
     *
     * @throws  UnknownOptionException  If an option hasn't been defined.
     * @throws  MissingParameterException       If a required parameter is
     *                  missing.
     * @throws  UnexpectedParameterException    If an unexpected parameter is
     *                  encountered.
     */
    public int parse(String[] args)
         throws UnknownOptionException,
                MissingParameterException,
                UnexpectedParameterException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "parse(String[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "args: " + args);
        }

        result = new Hashtable ();

        // First parse the command line.
        int  i;
        for (i = 0; i < args.length; i++) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "parse(String[])",
                           "args[" + i + "]: " + args[i]);
            }
            if (args[i].startsWith("-")) {
                if (args[i].equals("--")) {
                    // End of options
                    i++;
                    break;
                } else if (args[i].indexOf("=") > 0) {
                    // Name/value pair in one option
                    int pos = args[i].indexOf("=");
                    String option = args[i].substring(1, pos);
                    String value = args[i].substring(pos + 1);
                    addResult(isAllowed(option), option, value);
                } else {
                    // normal option
                    String option = args[i].substring(1);
                    Option opt = isAllowed(option);
                    if (opt == null) {
                        throw new UnknownOptionException
                            ("Unknown option " + option, option);
                    }
                    String value = null;
                    if (args.length > i + 1
                        && opt.parameter != '-'
                        && opt.parameter != '='
                        && (args[i + 1].length() == 0
                            || args[i + 1].charAt(0) != '-')) {
                                value =  args[++i];
                    }
                    addResult(opt, option, value);
                }
            } else {
                break;
            }
        }

        // Check for extra options
        Object prop = propFile;
        if (property != null) {
            prop = System.getProperties().getProperty(property);
        }
        if (propOption != null && result.containsKey(propOption)) {
            prop = result.get(propOption);
        }
        if (prop instanceof String) {
            addProperties((String) prop);
        } else if (prop instanceof String[]) {
            for (int j = 0; j < ((String[]) prop).length; j++) {
                addProperties(((String[])prop)[j]);
            }
        }

        if (logger.level >= Log.DBGDETAIL) {
            logger.log(Log.DBGDETAIL, "parse(String[])",
                       "Options follow:");
            for (Enumeration keys = result.keys();
                 keys.hasMoreElements();) {
                String key = (String) keys.nextElement();
                Object value = result.get(key);
                if (value instanceof String) {
                    logger.log(Log.DBGDETAIL, "parse(String[])",
                               key + "=\"" + value + "\"");
                } else {
                    StringBuffer sb = new StringBuffer ("[");
                    for (int j = 0; j < ((String[]) value).length; j++) {
                        if (j > 0) {
                            sb.append(",");
                        }
                        sb.append(((String[]) value)[j]);
                    }
                    sb.append("]");
                    logger.log(Log.DBGDETAIL, "parse(String[])",
                               key + "=\"" + sb.toString() + "\"");
                }
            }
        }

        return i;
    }

    // }}}
    // {{{ getOptions

    /**
     * Get the parsed command line options. The keys are the options and the
     * values are their parameters, where the empty String can either mean no
     * parameter or an empty one. For options with multiple parameters the
     * value is a String array.
     *
     * @return  The parsed options.
     */
    public Hashtable getOptions()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getOptions()", "");
        }
        return result;
    }


    // }}}
    // {{{ getString

    /**
     * Get a String value from the parsed options.
     *
     * @param   name    The name of the option to look up.
     * @param   deflt   A default value in case the option is missing or the
     *                  value is not a String.
     *
     * @return  The option value or the default.
     *
     * @since   0.98.0
     */
    public String getString(String name, String deflt)
    {
        if (result == null) {
            return deflt;
        }
        Object ret = result.get(name);
        return (ret instanceof String) ? (String) ret : deflt;
    }

    // }}}
    // {{{ getStrings

    /**
     * Get a String array value from the parsed options.
     *
     * @param   name    The name of the option to look up.
     * @param   deflt   A default value in case the option is missing or the
     *                  value is not a String array.
     *
     * @return  The option value or the default.
     *
     * @since   0.98.0
     */
    public String[] getStrings(String name, String[] deflt)
    {
        if (result == null) {
            return deflt;
        }
        Object ret = result.get(name);
        return (ret instanceof String[]) ? (String[]) ret : deflt;
    }

    // }}}
    // {{{ getInt

    /**
     * Get an int value from the parsed options.
     *
     * @param   name    The name of the option to look up.
     * @param   deflt   A default value in case the option is missing or the
     *                  value is not a Number.
     *
     * @return  The option value or the default.
     *
     * @since   0.98.0
     */
    public int getInt(String name, int deflt)
    {
        if (result == null) {
            return deflt;
        }
        Object ret = result.get(name);
        if (ret instanceof String) {
            try {
                return Integer.parseInt((String) ret);
            } catch (NumberFormatException ex) {
            }
        }
        return deflt;
    }

    // }}}
    // {{{ getBoolean

    /**
     * Get a boolean value from the parsed options.
     *
     * @param   name    The name of the option to look up.
     * @param   deflt   A default value in case the option is missing.
     *
     * @return  The option value or the default. This is a special case of
     *          a String to boolean conversion, since the empty String also
     *          means true.
     *
     * @since   0.98.0
     */
    public boolean getBoolean(String name, boolean deflt)
    {
        if (result == null) {
            return deflt;
        }
        Object ret = result.get(name);
        if (ret instanceof String) {
            // Special case: empty String means true
            return ((String) ret).length() == 0
                || new Boolean((String) ret).booleanValue();
        }
        return deflt;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helpers
    //----------------------------------------------------------------------
    // {{{ isAllowed

    /**
     * Check whether an argument can be recognized as an option and return
     * the relevant Option if found.
     *
     * @param   name    The name from the command line.
     *
     * @return  The Option found or null if not valid.
     */
    private Option isAllowed(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isAllowed(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }

        // First check full option name
        Option ret = (Option) options.get(name);
        if (ret != null) {
            return ret;
        }

        // Then search for the best match among wildcarded options
        int best = 0;
        for (int i = 0; i < wildcards.size(); i++) {
            Option opt = (Option) wildcards.elementAt(i);
            if (opt.name.length() > best && name.startsWith(opt.name)) {
                ret = opt;
                best = opt.name.length();
            }
        }

        return ret;
    }

    // }}}
    // {{{ addResult

    /**
     * Add one option/parameter pair to the result.
     *
     * @param   opt     The Option defining the option.
     * @param   name    The name of the option.
     * @param   value   The parameter value or null.
     *
     * @throws  UnknownOptionException  If an option hasn't been defined.
     * @throws  MissingParameterException       If a required parameter is
     *                  missing.
     * @throws  UnexpectedParameterException    If an unexpected parameter is
     *                  encountered.
     */
    private void addResult(Option opt, String name, String value)
        throws UnknownOptionException,
               MissingParameterException,
               UnexpectedParameterException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addResult(String,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }

        // check for validity
        if (opt == null) {
            throw new UnknownOptionException
                ("Unknown option " + name, name);
        }
        if (opt.parameter == '-' && value != null) {
            throw new UnexpectedParameterException
                ("Illegal parameter " + value + " for option " + name,
                 name, value);
        }
        if (opt.parameter == '+' && value == null) {
            throw new MissingParameterException
                ("Missing parameter for option " + name, name);
        }

        // Make value Hashtable compatible
        if (value == null) {
            value = "";
        }

        // Create the result
        if (opt.multiple) {
            // Need a String array
            String[] arr;
            if (result.containsKey(name)) {
                String[] old = (String[]) result.get(name);
                arr = new String[old.length + 1];
                System.arraycopy(old, 0, arr, 0, old.length);
                arr[old.length] = value;
            } else {
                arr = new String[] { value };
            }
            result.put(name, arr);
        } else {
            result.put(name, value);
        }
    }

    // }}}
    // {{{ addProperties

    /**
     * Add properties from a file or URL to the result options.
     *
     * @param   prop    The file name or URL of the properties to add.
     *
     * @throws  UnknownOptionException  If an option hasn't been defined.
     * @throws  MissingParameterException       If a required parameter is
     *                  missing.
     * @throws  UnexpectedParameterException    If an unexpected parameter is
     *                  encountered.
     */
    private void addProperties(String prop)
        throws UnknownOptionException,
               MissingParameterException,
               UnexpectedParameterException
    {
        InputStream in = null;
        try {
            // first try as file
            in = new FileInputStream (prop);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "parse(String[])",
                           "Reading properties from file " + prop);
            }
        } catch (IOException ex) {
            if (logger.level >= Log.MSG) {
                logger.log(Log.MSG, "parse(String[])", ex);
            }
            try {
                // then try as URL
                in = new URL (prop).openStream();
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "parse(String[])",
                               "Reading properties from URL " + prop);
                }
            } catch (IOException ex2) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "parse(String[])", ex2);
                }
            }
        }
        if (in != null) {
            try {
                // Load properties
                Properties props = new Properties ();
                props.load(in);
                // Override with options
                for (Enumeration keys = props.keys();
                     keys.hasMoreElements();) {
                    String key = (String) keys.nextElement();
                    Option opt = isAllowed(key);
                    if (opt != null &&
                        (opt.multiple || !result.containsKey(key))) {
                        String val = props.getProperty(key);
                        if (val.length() == 0 && opt.parameter == '-') {
                            val = null;
                        }
                        addResult(opt, key, val);
                    }
                }
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("parse(String[])", ex);
                }
            }
        }
    }

    // }}}

    // {{{ class Option

    /**
     * An Option is a simple struct holding the attributes needed to parse an
     * argument list.
     */
    private static class Option
    {
        /**
         * The name of the Option.
         */
        public String name;

        /**
         * Whether parameters are required.
         */
        public char parameter;

        /**
         * Whether multiple values are allowed.
         */
        public boolean multiple;

        /**
         * Whether similar Options are allowed.
         */
        public boolean wildcard;

        /**
         * Create a new Option.
         *
         * @param       name            The name of the Option.
         * @param       parameter       Whether parameters are required.
         * @param       multiple        Whether multiple values are allowed.
         * @param       wildcard        Whether similar Options are allowed.
         */
        public Option(String name, char parameter, boolean multiple,
                      boolean wildcard)
        {
            this.name = name;
            this.parameter = parameter;
            this.multiple = multiple;
            this.wildcard = wildcard;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------

    // {{{ class UnitTest

    /**
     * Test cases for the ArgsParser class.
     */
    public static class UnitTest extends TestCase
    {
        // {{{ variables

        /**
         * The parser to test.
         */
        private ArgsParser parser;

        // }}}

        // {{{ constructor

        /**
         * Create a new UnitTest.
         *
         * @param       name    The name of the test.
         */
        public UnitTest(String name)
        {
                super(name);
        }

        // }}}

        // {{{ setUp

        /**
         * Set up the test fixture.
         */
        protected void setUp()
        {
            //  Log.setOutputLevel(10);
            //  Log.setPreQueueLevel(10);
            //  Log.setQueueing(false);
            //  Logger.setDefaultLogLevel(10);
            parser = new ArgsParser ();
        }

        // }}}
        // {{{ suite

        /**
         * Get the suite of tests.
         *
         * @return      The tests.
         */
        public static Test suite()
        {
                return new TestSuite(UnitTest.class);
        }

        // }}}

        // {{{ testBadOptions

        /**
         * Test bad options in setOptions method.
         */
        public void testBadOptions()
        {
            // bad options
            try {
                parser.setOptions(new String[] {null});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }

            try {
                parser.setOptions(new String[] {""});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }

            try {
                parser.setOptions(new String[] {"+"});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }

            try {
                parser.setOptions(new String[] {"m+"});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }

            try {
                parser.setOptions(new String[] {"mmm"});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }


            try {
                parser.setOptions(new String[] {"aa"});
                assert(false);
            } catch (IllegalArgumentException ex) {
            }

            // OK options
            parser.setOptions(new String[] {
                "m+*", "+*", "m-a*", "+x"});
        }

        // }}}
        // {{{ testParseOK

        /**
         * Test the parse method.
         */
        public void testParseOK()
        {
            parser.setOptions(new String[] {"+a", "-b", "?c", "=d"});

            try {
                parser.parse(new String[] {"-a", "1", "-c", "-b" , "-d"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            Hashtable result = parser.getOptions();
            assert(result.size() == 4);
            assert(result.get("a").equals("1"));
            assert(result.get("b").equals(""));
            assert(result.get("c").equals(""));
            assert(result.get("d").equals(""));

            try {
                parser.parse(new String[] {"-a", "1", "-c", "2", "-d", "stop",
                                           "-b"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            result = parser.getOptions();
            assert(result.size() == 3);
            assert(result.get("a").equals("1"));
            assertNull(result.get("b"));
            assert(result.get("c").equals("2"));
            assert(result.get("d").equals(""));

            try {
                parser.parse(new String[] {"-a", "1", "-c", "2", "-d=stop",
                                           "-b"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            result = parser.getOptions();
            assert(result.size() == 4);
            assert(result.get("a").equals("1"));
            assert(result.get("b").equals(""));
            assert(result.get("c").equals("2"));
            assert(result.get("d").equals("stop"));
        }

        // }}}
        // {{{ testParseBad

        /**
         * Test the parse method.
         */
        public void testParseBad()
        {
            parser.setOptions(new String[] {"+a", "-b", "?c"});

            try {
                parser.parse(new String[] {"-a", "-c", "-b"});
                assert(false);
            } catch (MalformedArgumentListException ex) {
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "testParseBad()", ex);
                }
            }

            try {
                parser.parse(new String[] {"-a", "1", "-c", "2", "-b=2"});
                assert(false);
            } catch (MalformedArgumentListException ex) {
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "testParseBad()", ex);
                }
            }
        }

        // }}}
        // {{{ testParseMultiple

        /**
         * Test the parse method.
         */
        public void testParseMultiple()
        {
            parser.setOptions(new String[] {"m+a", "m-b", "m?c"});

            try {
                parser.parse(new String[] {"-a", "1", "-a", "",
                                           "-c", "-c=-b","-b", "-b"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            Hashtable result = parser.getOptions();
            assert(result.size() == 3);

            String[] arr = (String[]) result.get("a");
            assert(arr.length == 2);
            assert(arr[0].equals("1"));
            assert(arr[1].equals(""));

            arr = (String[]) result.get("b");
            assert(arr[0].equals(""));
            assert(arr[1].equals(""));

            arr = (String[]) result.get("c");
            assert(arr[0].equals(""));
            assert(arr[1].equals("-b"));

        }

        // }}}
        // {{{ testParseWildcards

        /**
         * Test the parse method.
         */
        public void testParseWildcards()
        {
            parser.setOptions(new String[] {"+a*"});

            try {
                parser.parse(new String[] {"-a", "1", "-aa", "2"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            Hashtable result = parser.getOptions();
            assert(result.size() == 2);
            assert(result.get("a").equals("1"));
            assert(result.get("aa").equals("2"));

            try {
                parser.parse(new String[] {"-a", "1", "-c", "2"});
                assert(false);
            } catch (MalformedArgumentListException ex) {
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "testParseBad()", ex);
                }
            }

            try {
                parser.parse(new String[] {"-a", "-aa"});
                assert(false);
            } catch (MalformedArgumentListException ex) {
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "testParseBad()", ex);
                }
            }

            parser.setOptions(new String[] {"m+a*"});
            try {
                parser.parse(new String[] {"-aa", "1", "-aa", "2"});
            } catch (MalformedArgumentListException ex) {
                ex.printStackTrace();
                assert(false);
            }
            result = parser.getOptions();
            assert(result.size() == 1);
            String[] arr = (String[]) result.get("aa");
            assert(arr.length == 2);
            assert(arr[0].equals("1"));
            assert(arr[1].equals("2"));
        }

        // }}}
    }

    // }}}

}
