// {{{ copyright

/********************************************************************
 *
 * $Id: DynamicClassLoader.java,v 1.11 2001/02/15 18:06:18 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 * Florian Rodler <rodler@in.tum.de>
 *   - noted missing InputStream close in loadClass(String,boolean)
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;

import java.lang.reflect.InvocationTargetException;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * A DynamicClassLoader is a valuable tool, particularly during the
 * development phase of a program, since it can load new versions of a class
 * into a running program without restarting it. There are a few restrictions
 * though: <p>
 *
 * If a class file is older than the initialization time of the program, the
 * standard system class will be used. If it is newer, a new
 * DynamicClassLoader will be used to load and define the class, since any
 * ClassLoader can hold only one class of the same name. <p>
 *
 * Other classes used by the newly created class will be resolved through the
 * same DynamicClassLoader, but unless they have also been modified recently,
 * their system class versions will be used. In this way it is possible to
 * cast an object of a dynamically reinstantiated class to its system base
 * class or interface and use it accordingly, i.e.
 *
 * <pre>
 * JFrame frame = (JFrame) DynamicClassLoader.getObject("my.package.MySpecialFrame");
 * </pre>
 *
 * will work, while
 *
 * <pre>
 * JFrame frame = (MySpecialFrame) DynamicClassLoader.getObject("my.package.MySpecialFrame");
 * </pre>
 *
 * will only work until the MySpecialFrame.class file is modified and a
 * DynamicClassLoader is used to create a new version of the class, since the
 * system class and the dynamic class are incompatible for assignment. <p>
 *
 * If you want to reload a class whose (direct or indirect) base class has
 * been changed you have to recompile the class itself as well, otherwise
 * its cached version is used, which refers to the old version of the base
 * class.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.11 $
 */

// }}}
public class DynamicClassLoader
    extends ClassLoader
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (DynamicClassLoader.class);

    /**
     * Cache for system classes.
     */
    protected static Hashtable systemClasses = new Hashtable ();

    /**
     * Cache for loaded classes.
     */
    protected static Hashtable loadedClasses = new Hashtable ();

    /**
     * Cache for files for class resources.
     */
    protected static Hashtable files = new Hashtable ();

    /**
     * Special File to cache unknowns.
     */
    protected static final File fileNotFound = new File ("file not found");

    /**
     * Files and ZipFiles for the classpath.
     */
    protected static Vector classpath;

    /**
     * Initialization time. Any non-system class file newer than this will be
     * dynamically created.
     */
    protected static long initTime = System.currentTimeMillis();

    /**
     * The primary class loader that will be used to load classes that have
     * changed for the first time after initialization. For every further
     * change of the class file a new DynamicClassLoader will be used.
     */
    protected static DynamicClassLoader primaryLoader =
        new DynamicClassLoader();

    /**
     * Cache for dynamically created classes.
     */
    protected Hashtable classes = new Hashtable ();

    // }}}

    // {{{ constructor

    /**
     * Create a new DynamicClassLoader.
     */
    public DynamicClassLoader ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "DynamicClassLoader()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Static convenience methods
    //----------------------------------------------------------------------
    // {{{ initClasspath

    /**
     * Initialize the classpath. All elements of the classpath are looked up
     * and either stored as Files or ZipFiles. This method is called
     * automatically but you may have to call it again if you modify the
     * <code>java.class.path</code> system property. <p>
     *
     * Since qflib 0.98.0 also handles the lib/ext directory of JRE 1.2 and
     * above.
     */
    public static void initClasspath()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "initClasspath()", "");
        }
        // clear file cache
        files = new Hashtable ();
        classpath = new Vector ();
        String cp = System.getProperty("java.class.path");
        if (cp != null) {
            StringTokenizer st = new StringTokenizer (cp, File.pathSeparator);
            for (; st.hasMoreTokens();) {
                File path = new File (st.nextToken());

                if (path.exists()) {
                    if (path.isDirectory()) {
                        classpath.addElement(path);
                    } else {
                        try {
                            classpath.addElement(new ZipFile (path));
                        } catch (ZipException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("initClasspath()", ex);
                            }
                        } catch (IOException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("initClasspath()", ex);
                            }
                        }
                    }
                }
            }
        }

        String extdirs = System.getProperty("java.ext.dirs");
        if (extdirs == null) {
            return;
        }

        StringTokenizer st = new StringTokenizer (extdirs, File.pathSeparator);
        ExtensionFileFilter filter = new ExtensionFileFilter
            ("", new String[] {".jar", ".zip"}, false, true);
        for (; st.hasMoreTokens();) {
            File dir = new File (st.nextToken());
            if (dir.isDirectory()) {
                String[] jars = dir.list(filter);
                for (int i = 0; i < jars.length; i++) {
                    try {
                        classpath.addElement
                            (new ZipFile (new File (dir, jars[i])));
                    } catch (ZipException ex) {
                        if (logger.level >= Log.ERR) {
                            logger.log("initClasspath()", ex);
                        }
                    } catch (IOException ex) {
                        if (logger.level >= Log.ERR) {
                            logger.log("initClasspath()", ex);
                        }
                    }
                }
            }
        }
    }

    // }}}
    // {{{ loadDynamicClass

    /**
     * Get a Class object for a class name.
     *
     * @param   name    The fully qualified name of the class to get.
     *
     * @return  The Class object.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    public static Class loadDynamicClass(String name)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "loadDynamicClass(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }

        // check the primaryLoader's cache
        ClassEntry ce = (ClassEntry) loadedClasses.get(name);
        if (ce != null) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadDynamicClass(String)",
                           "Loaded classes cache hit");
            }
            if (ce.file.lastModified() <= ce.timestamp) {
                return ce.clazz;
            } else {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "loadDynamicClass(String)",
                               "Class file is new");
                }
                // The class has already been modified twice since the
                // program started. This is the only case where a new
                // DynamicClassLoader is needed.
                // This could be optimized by caching these classes also.
                return new DynamicClassLoader ().loadClass(name);
            }
        }

        return primaryLoader.loadClass(name);
    }

    // }}}
    // {{{ getDynamicObject

    /**
     * Get an Object instance of a class. The class must have a public zero
     * argument constructor.
     *
     * @param   classname       The fully qualified class name.
     *
     * @return  The new instance.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     * @throws  InstantiationException  If the class cannot be instantiated.
     * @throws  IllegalAccessException  If the caller doesn't have access to
     *          the class.
     */
    public static Object getDynamicObject(String classname)
         throws ClassNotFoundException,
             InstantiationException,
             IllegalAccessException
    {
        return loadDynamicClass(classname).newInstance();
    }

    // }}}
    // {{{ getDynamicObject

    /**
     * Get an Object instance of a class. The class must have a constructor
     * matching arguments.
     *
     * @param   classname       The fully qualified class name.
     * @param   types           The argument classes for the constructor.
     * @param   arguments       The arguments for the constructor.
     *
     * @return  The new instance.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     * @throws  InstantiationException  If the class cannot be instantiated.
     * @throws  IllegalAccessException  If the caller doesn't have access to
     *          the class.
     * @throws  NoSuchMethodException   If no matching constructor exists.
     * @throws  InvocationTargetException If the constructor throws an
     *          exception.
     *
     * @see     java.lang.Class#getConstructor(Class[])
     *
     * @since   0.96.0
     */
    public static Object getDynamicObject(String classname,
                                          Class[] types,
                                          Object[] arguments)
         throws ClassNotFoundException,
             InstantiationException,
             IllegalAccessException,
             NoSuchMethodException,
             InvocationTargetException
    {
        Class cl = loadDynamicClass(classname);
        return cl.getConstructor(types).newInstance(arguments);
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ loadClass

    /**
     * Load and resolve the class named name.
     *
     * @param   name    The class name.
     *
     * @return  The class.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    public Class loadClass(String name)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "loadClass(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        return loadClass(name, true);
    }

    // }}}
    // {{{ getResourceAsStream

    /**
     * Find a resource in the classpath and open a stream for it.
     *
     * @param   name    The name of the resource.
     *
     * @return  The stream.
     */
    public InputStream getResourceAsStream(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getResourceAsStream(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }

        // check the cache
        File f = (File) files.get(name);
        if (f != null) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "getResourceAsStream(String)",
                           "Cache hit");
            }
            if (f.equals(fileNotFound)) {
                return null;
            }
            try {
                return new FileInputStream (f);
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "getResourceAsStream(String)", ex);
                }
                return null;
            }
        }

        for (int i = 0; i < classpath.size(); i++) {
            Object path = classpath.elementAt(i);

            if (path instanceof File) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "getResourceAsStream(String)",
                               "Trying path " + path);
                }
                String sysName = name;
                if (!(File.separatorChar == '/')) {
                    sysName =
                        name.replace('/', File.separator
                                          .charAt(0));
                }
                try {
                    f = new File
                        (((File) path).getCanonicalPath() +
                         File.separator +
                         sysName);
                    if (! f.exists()) {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG,
                                       "getResourceAsStream(String)",
                                       f.toString() + "Not Found");
                        }
                        continue;
                    }
                    files.put(name, f);
                    return new FileInputStream (f);
                } catch (IOException ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log("getResourceAsStream(String)", ex);
                    }
                    return null;
                }
            } else if (path instanceof ZipFile) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "getResourceAsStream(String)",
                               "Trying ZipFile " +
                               ((ZipFile) path).getName());
                }
                try {
                    ZipEntry ze = ((ZipFile) path).getEntry(name);
                    if (ze == null) {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "getResourceAsStream(String)",
                                       "Not Found");
                        }
                        continue;
                    }
                    files.put(name, path);
                    return ((ZipFile) path).getInputStream(ze);
                } catch (ZipException ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log("getResourceAsStream(String)", ex);
                    }
                    return null;
                } catch (IOException ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log("getResourceAsStream(String)", ex);
                    }
                    return null;
                }
            }
        }
        files.put(name, fileNotFound);
        return null;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ loadClass

    /**
     * Load a class and maybe resolve it.
     *
     * @param   name    The class name.
     * @param   resolve Whether to resolve the class.
     *
     * @return  The class.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    protected synchronized Class loadClass(String name, boolean resolve)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "loadClass(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "resolve: " + resolve);
        }

        // check for system class
        if (isSystemClass(name)) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "System class " + name);
            }
            return makeSystemClass(name, resolve);
        }

        byte[] buf;
        Object source = null;
        File file = null;
        ZipFile zip = null;

        // check the private cache
        ClassEntry ce = (ClassEntry) classes.get(name);
        if (ce != null) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "Private cache hit");
            }
            if (ce.file.lastModified() > ce.timestamp) {
                // cannot reload the same class
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "loadClass(String,boolean)",
                               "Warning: class file is more recent than " +
                               "returned class");
                }
            }
            return ce.clazz;
        }

        InputStream is = null;
        // make sure the stream will be closed
        try {
            // check the system cache
            ce = (ClassEntry) systemClasses.get(name);
            if (ce != null) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "loadClass(String,boolean)",
                               "System cache hit");
                }
                if (ce.file.lastModified() <= ce.timestamp) {
                    return ce.clazz;
                } else {
                    // migrate from system to private cache
                    systemClasses.remove(name);
                    classes.put(name, ce);
                    loadedClasses.put(name, ce);
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "loadClass(String,boolean)",
                                   "Class file is new");
                    }
                    try {
                        if (ce.zip == null) {
                            is = new FileInputStream (ce.file);
                        } else {
                            ZipEntry ze = ce.zip.getEntry(name);
                            if (ze == null) {
                                if (logger.level >= Log.DBG) {
                                    logger.log(Log.DBG,
                                               "loadClass(String,boolean)",
                                               "Not Found in" +
                                               ce.zip.getName());
                                }
                                throw new ClassNotFoundException (name);
                            }
                            is = ce.zip.getInputStream(ze);
                        }
                    } catch (IOException ex) {
                        if (logger.level >= Log.ERR) {
                            logger.log("loadClass(String,boolean)", ex);
                        }
                        throw new ClassNotFoundException (name);
                    }
                }
            } else {
                String resName = name.replace('.', '/') + ".class";
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "loadClass(String,boolean)",
                               "Converted name is " + resName);
                }

                is = getResourceAsStream(resName);
                if (is == null) {
                    if (logger.level >= Log.MSG) {
                        logger.log(Log.MSG, "loadClass(String,boolean)",
                                   "Null resource for " + resName);
                    }
                    throw new ClassNotFoundException (name);
                }
                source = files.get(resName);
                if (source instanceof ZipFile) {
                    zip = (ZipFile) source;
                    file = new File (zip.getName());
                } else {
                    file = (File) source;
                }
                if (file.lastModified() <= initTime) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "loadClass(String,boolean)",
                                   "Old class file - use system class");
                    }
                    Class clazz = makeSystemClass(name, resolve);
                    if (zip != null) {
                        systemClasses.put(name, new ClassEntry (clazz, zip));
                    } else {
                        systemClasses.put(name, new ClassEntry (clazz, file));
                    }
                    return clazz;
                }
            }

            try {
                buf = StreamInput.read(is);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "loadClass(String,boolean)",
                               "class " + name + " has " +
                               buf.length + " bytes");
                }
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("loadClass(String,boolean)", ex);
                }
                throw new ClassNotFoundException (name);
            }
        } finally {
            // close the stream
            if (is != null) {
                try {
                    is.close();
                } catch (IOException ex) {
                    if (logger.level >= Log.WRN) {
                        logger.log(Log.WRN, "loadClass(String,boolean)", ex);
                    }
                }
            }
        }

        Class clazz = null;
        try {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "defining " + name);
            }
            clazz = defineClass(name, buf, 0, buf.length);
        } catch (Throwable ex) {
            if (logger.level >= Log.ERR) {
                logger.log("loadClass(String,boolean) " + name, ex);
            }
            throw new ClassNotFoundException (name);
        }
        if (ce == null) {
            if (zip != null) {
                ce = new ClassEntry (clazz, zip);
            } else {
                ce = new ClassEntry (clazz, file);
            }
            classes.put(name, ce);
            loadedClasses.put(name, ce);
        } else {
            ce.timestamp = ce.file.lastModified();
            ce.clazz = clazz;
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "loadClass(String,boolean)",
                       name + " defined");
        }
        if (resolve) {
            resolveClass(clazz);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           name + " resolved");
            }
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "loadClass(String,boolean)",
                       "finished, clazz: " + clazz +
                       " (" + clazz.hashCode() + ")");
        }
        return clazz;
    }

    // }}}
    // {{{ isSystemClass

    /**
     * Check whether a class name refers to a system class.
     *
     * @param   name    The name to check.
     *
     * @return  True if it is a system class.
     */
    protected static boolean isSystemClass(String name)
    {
        return (name.startsWith("java.")
                || name.startsWith("javax.")
                || name.startsWith("org.omg.")
                || name.startsWith("com.sun."));
    }

    // }}}
    // {{{ makeSystemClass

    /**
     * Get a system class and maybe resolve it.
     *
     * @param   name    The class name.
     * @param   resolve Whether to resolve the class.
     *
     * @return  The class.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    protected Class makeSystemClass (String name, boolean resolve)
         throws ClassNotFoundException
    {
        Class clazz = findSystemClass(name);
        if (clazz == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "makeSystemClass(String,boolean)",
                           "Cannot fing system class " + name);
            }
            throw new ClassNotFoundException (name);
        }
        if (resolve) {
            resolveClass(clazz);
        }
        return clazz;
    }

    // }}}

    static {
        initClasspath();
    }
}
