// {{{ copyright

/********************************************************************
 *
 * $Id: RelocatingObjectInputStream.java,v 1.5 2000/06/06 16:41:29 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectStreamClass;
import java.io.IOException;
import java.io.StreamCorruptedException;

// }}}

/**
 * The RelocatingObjectInputStream is useful when a class has been moved to a
 * different package and some serialized Objects of its type should be
 * restored. Simply pass two arrays of Strings to the constructor, which map
 * the names of old packages to new packages. The serialVersionUID of
 * the old an new class must still match though.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 */
public class RelocatingObjectInputStream extends ObjectInputStream
{
    // {{{ variables

    /**
     * Package names to convert.
     */
    private String[] from;

    /**
     * Converted package names.
     */
    private String[] to;

    // }}}

    // {{{ constructor

    /**
     * Create a new RelocatingObjectInputStream. The arrays of package names
     * must be of equal length. For each deserialized class the first name
     * that matches the start of the class name will be replaced by the
     * corresponding name from the second list.
     *
     * @param   in      The ObjectInputStream to extend.
     * @param   from    Names of old packages.
     * @param   to      Names of the new packages.
     *
     * @throws  IOException     If an exception occurred in the underlying
     *                          stream.
     * @throws  StreamCorruptedException        If the version or magic number
     *                                          are incorrect.
     */
    public RelocatingObjectInputStream(InputStream in, String[] from,
                                       String[] to)
         throws IOException, StreamCorruptedException
    {
        super(in);
        if (from.length != to.length) {
            throw new IllegalArgumentException
                ("from and to differ in length");
        }
        this.from = from;
        this.to = to;
    }

    // }}}

    // {{{ resolveClass

    /**
     * Load the local class equivalent of the specified stream class
     * description. Subclasses may implement this method to allow classes to
     * be fetched from an alternate source. The corresponding method in
     * ObjectOutputStream is annotateClass. This method will be invoked only
     * once for each unique class in the stream. This method can be
     * implemented by subclasses to use an alternate loading mechanism but
     * must return a Class object. Once returned, the serialVersionUID of the
     * class is compared to the serialVersionUID of the serialized class. If
     * there is a mismatch, the deserialization fails and an exception is
     * raised.  By default the class name is resolved relative to the class
     * that called readObject.
     *
     * @param   osc     The ObjectStreamClass describing the class.
     *
     * @return  The resolved class.
     *
     * @throws  IOException     If an I/O error occurs.
     * @throws  ClassNotFoundException  If class of a serialized object cannot
     *                                  be found.
     */
    protected Class resolveClass(ObjectStreamClass osc)
         throws IOException, ClassNotFoundException
    {
        String name = osc.getName();
        for (int i = 0; i < from.length; i++) {
            String array;
            for (array = ""; name.startsWith("[L");
                 name = name.substring(2)) {
                array = array + "[L";
            }
            if (name.startsWith(from[i])) {
                name = array + to[i] + name.substring(from[i].length());
                return Class.forName(name);
            }
        }
        return super.resolveClass(osc);
    }

    // }}}
}
