// {{{ copyright

/********************************************************************
 *
 * $Id: ResourceWaiter.java,v 1.3 2000/06/06 16:41:29 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This class is useful in cases where several Threads are waiting for a
 * limited number of resources. Instead of simply calling wait, a
 * ResourceWaiter can be pushed onto a queue.  When a resource becomes
 * available, the next waiter is popped from the queue and handed the
 * resource, so no other Thread can grab it first. <p>
 *
 * The {@link ThreadPool ThreadPool} is an example for a class that makes use
 * of a ResourceWaiter.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.3 $
 */
public class ResourceWaiter
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (ResourceWaiter.class);

    /**
     * The resource passed to the ResourceWaiter.
     */
    private Object object;

    /**
     * Signal that waiting is over.
     */
    private boolean done;

    /**
     * The lock to synchronize on.
     */
    private Object lock;

    // }}}

    // {{{ constructors

    /**
     * Create a new ResourceWaiter
     */
    public ResourceWaiter ()
    {
        lock = this;
    }

    /**
     * Create a new ResourceWaiter
     *
     * @param   lock    The object to wait on. When using a ResourceWaiter from
     *                  a synchronized method, make sure to use the same lock
     *                  for the ResourceWaiter that the caller owns, otherwise
     *                  deadlocks will occur.
     */
    public ResourceWaiter (Object lock)
    {
        this.lock = lock;
    }

    // }}}

    // {{{ waitFor

    /**
     * Wait until either someone hands an Object to the waiter or a timeout
     * is reached.
     *
     * @param   timeout The maximum time to wait in milliseconds. Negative
     *          means wait indefinitely.
     *
     * @return  The Object passed to gotIt or null, if the timeout was reached.
     */
    public Object waitFor(long timeout)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "waitFor(long)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "timeout: " + timeout);
        }

        long start = 0;
        long now = 0;

        if (timeout > 0) {
            start = System.currentTimeMillis();
            now = start;
        }

        synchronized (lock) {
            while (!done) {
                if (timeout >= 0) {
                    if (now >= start + timeout) {
                        return null;
                    }
                    try {
                        lock.wait(start + timeout - now);
                    } catch (InterruptedException ex) {
                    }
                    now = System.currentTimeMillis();
                } else {
                    try {
                        lock.wait();
                    } catch (InterruptedException ex) {
                    }
                }
            }
        }
        return object;
    }


    // }}}
    // {{{ handResource

    /**
     * Pass an Object to the ResourceWaiter.
     *
     * @param   object  The Object that the ResourceWaiter waited for. This
     *                  will cause the waitFor method of the ResourceWaiter
     *                  to continue and hand the object back to its caller.
     */
    public synchronized void handResource(Object object)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "handResource(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "object: " + object);
        }
        synchronized (lock) {
            done = true;
            this.object = object;
            lock.notifyAll();
        }
    }

    // }}}
}
